# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Functionality to build extensible variable spaces into ReFrame tests.
#

import copy
import math

import reframe.core.fields as fields
import reframe.core.namespaces as namespaces
from reframe.core.exceptions import ReframeSyntaxError
from reframe.core.warnings import (user_deprecation_warning,
                                   suppress_deprecations)


class _UndefinedType:
    '''Custom type to flag a variable as undefined.'''
    __slots__ = ()

    def __deepcopy__(self, memo):
        return self


Undefined = _UndefinedType()

DEPRECATE_RD = 1
DEPRECATE_WR = 2
DEPRECATE_RDWR = DEPRECATE_RD | DEPRECATE_WR


class TestVar:
    '''Insert a new  test variable.

    Declaring a test variable through the :func:`variable` built-in allows for
    a more robust test implementation than if the variables were just defined
    as regular test attributes (e.g. ``self.a = 10``). Using variables
    declared through the :func:`variable` built-in guarantees that these
    regression test variables will not be redeclared by any child class, while
    also ensuring that any values that may be assigned to such variables
    comply with its original declaration. In essence, declaring test variables
    with the :func:`variable` built-in removes any potential test errors that
    might be caused by accidentally overriding a class attribute. See the
    example below.

    .. code:: python

       class Foo(rfm.RegressionTest):
           my_var = variable(int, value=8)
           not_a_var = my_var - 4

           @run_after('init')
           def access_vars(self):
               print(self.my_var) # prints 8.
               # self.my_var = 'override'  # Error: my_var must be an int!
               self.not_a_var = 'override' # This will work, but is dangerous!
               self.my_var = 10 # tests may also assign values the standard way

    Here, the argument ``value`` in the :func:`variable` built-in sets the
    default value for the variable. This value may be accessed directly from
    the class body, as long as it was assigned before either in the same class
    body or in the class body of a parent class. This behavior extends the
    standard Python data model, where a regular class attribute from a parent
    class is never available in the class body of a child class. Hence, using
    the :func:`variable` built-in enables us to directly use or modify any
    variables that may have been declared upstream the class inheritance
    chain, without altering their original value at the parent class level.

    .. code:: python

       class Bar(Foo):
           print(my_var) # prints 8
           # print(not_a_var) # This is standard Python and raises a NameError

           # Since my_var is available, we can also update its value:
           my_var = 4

           # Bar inherits the full declaration of my_var with the original
           # type-checking.
           # my_var = 'override' # Wrong type error again!

           @run_after('init')
           def access_vars(self):
               print(self.my_var) # prints 4
               print(self.not_a_var) # prints 4


       print(Foo.my_var) # prints 8
       print(Bar.my_var) # prints 4


    Here, :class:`Bar` inherits the variables from :class:`Foo` and can see
    that ``my_var`` has already been declared in the parent class. Therefore,
    the value of ``my_var`` is updated ensuring that the new value complies to
    the original variable declaration. However, the value of ``my_var`` at
    :class:`Foo` remains unchanged.

    These examples above assumed that a default value can be provided to the
    variables in the bases tests, but that might not always be the case. For
    example, when writing a test library, one might want to leave some
    variables undefined and force the user to set these when using the test.
    As shown in the example below, imposing such requirement is as simple as
    not passing any ``value`` to the :func:`variable` built-in, which marks
    the given variable as *required*.

    .. code:: python

       # Test as written in the library
       class EchoBaseTest(rfm.RunOnlyRegressionTest):
           what = variable(str)

           valid_systems = ['*']
           valid_prog_environs = ['*']

         @run_before('run')
         def set_executable(self):
             self.executable = f'echo {self.what}'

         @sanity_function
         def assert_what(self):
             return sn.assert_found(fr'{self.what}')


       # Test as written by the user
       @rfm.simple_test
       class HelloTest(EchoBaseTest):
           what = 'Hello'


       # A parameterized test with type-checking
       @rfm.simple_test
       class FoodTest(EchoBaseTest):
           param = parameter(['Bacon', 'Eggs'])

         @run_after('init')
         def set_vars_with_params(self):
             self.what = self.param


    Similarly to a variable with a value already assigned to it, the value of
    a required variable may be set either directly in the class body, on the
    :func:`__init__` method, or in any other hook before it is referenced.
    Otherwise an error will be raised indicating that a required variable has
    not been set. Conversely, a variable with a default value already assigned
    to it can be made required by assigning it the ``required`` keyword.
    However, this ``required`` keyword is only available in the class body.

    .. code:: python

       class MyRequiredTest(HelloTest):
         what = required


    Running the above test will cause the :func:`set_exec_and_sanity` hook
    from :class:`EchoBaseTest` to throw an error indicating that the variable
    ``what`` has not been set.

    Finally, variables may alias each other. If a variable is an alias of
    another one it behaves in the exact same way as its target. If a change is
    made to the target variable, this is reflected to the alias and vice
    versa. However, alias variables are independently loggable: an alias may
    be logged but not its target and vice versa. Aliased variables are useful
    when you want to rename a variable and you want to keep the old one for
    compatibility reasons.

    :param `types`: the supported types for the variable.
    :param value: the default value assigned to the variable. If no value is
        provided, the variable is set as ``required``.
    :param field: the field validator to be used for this variable. If no
        field argument is provided, it defaults to
        :attr:`reframe.core.fields.TypedField`. The provided field validator
        by this argument must derive from :attr:`reframe.core.fields.Field`.
    :param alias: the target variable if this variable is an alias. This must
        refer to an already declared variable and neither default value nor a
        field can be specified for an alias variable.
    :param loggable: Mark this variable as loggable. If :obj:`True`, this
        variable will become a log record attribute under the name
        ``check_NAME``, where ``NAME`` is the name of the variable.
    :param `kwargs`: keyword arguments to be forwarded to the constructor of
        the field validator.
    :returns: A new test variable.

    .. versionadded:: 3.10.2
       The ``loggable`` argument is added.

    .. versionadded:: 4.0.0
       Alias variable are introduced.

    '''

    # NOTE: We can't use truly private fields in `__slots__`, because
    # `__setattr__()` will be called with their mangled name and we cannot
    # match them in the `__slots__` without making implementation-defined
    # assumptions about the mangled name. So we just add the `_p_` prefix for
    # to denote the "private" fields.

    __slots__ = ('_p_default_value', '_p_field',
                 '_loggable', '_name', '_target')

    __mutable_props = ('_default_value',)

    def __init__(self, *args, **kwargs):
        alias = kwargs.pop('alias', None)
        if alias is not None and 'field' in kwargs:
            raise ValueError(f"'field' cannot be set for an alias variable")

        if alias is not None and 'value' in kwargs:
            raise ValueError('alias variables do not accept default values')

        if alias is not None and not isinstance(alias, TestVar):
            raise TypeError(f"'alias' must refer to a variable; "
                            f"found {type(alias).__name__!r}")

        field_type = kwargs.pop('field', fields.TypedField)
        if alias is not None:
            self._p_default_value = alias._default_value
        else:
            self._p_default_value = kwargs.pop('value', Undefined)

        self._loggable = kwargs.pop('loggable', False)
        if not issubclass(field_type, fields.Field):
            raise TypeError(
                f'field {field_type!r} is not derived from '
                f'{fields.Field.__qualname__}'
            )

        if alias is not None:
            self._p_field = alias._field
        else:
            self._p_field = field_type(*args, **kwargs)

        self._target = alias

    @classmethod
    def create_deprecated(cls, var, message,
                          kind=DEPRECATE_RDWR, from_version='0.0.0'):
        ret = TestVar.__new__(TestVar)
        ret._p_field = fields.DeprecatedField(var.field, message,
                                              kind, from_version)
        ret._p_default_value = var._default_value
        ret._loggable = var._loggable
        ret._target = var._target
        return ret

    def _warn_deprecation(self, kind):
        if self.is_deprecated() and self.field.op & kind:
            user_deprecation_warning(self.field.message)

    def is_deprecated(self):
        return isinstance(self._p_field, fields.DeprecatedField)

    def is_loggable(self):
        return self._loggable

    def is_defined(self):
        return self._default_value is not Undefined

    def is_alias(self):
        return self._target is not None

    def undefine(self):
        self._default_value = Undefined

    def define(self, value):
        self._warn_deprecation(DEPRECATE_WR)
        self._default_value = value

    @property
    def _default_value(self):
        if self.is_alias():
            return self._target._default_value
        else:
            return self._p_default_value

    @_default_value.setter
    def _default_value(self, value):
        if self.is_alias():
            self._target._default_value = value
        else:
            self._p_default_value = value

    @property
    def default_value(self):
        # Variables must be returned by-value to prevent an instance from
        # modifying the class variable space.
        self._check_is_defined()
        self._warn_deprecation(DEPRECATE_RD)
        return copy.deepcopy(self._default_value)

    @property
    def _field(self):
        if self.is_deprecated():
            return self._p_field

        if self._target:
            return self._target._field
        else:
            return self._p_field

    @property
    def field(self):
        return self._field

    @property
    def name(self):
        return self._name

    @property
    def target(self):
        return self._target

    def reset_target(self, new_target):
        if not self.is_deprecated():
            self._p_field = new_target._field
        else:
            self._p_field._target_field = new_target._field

        self._target = new_target

    def __set_name__(self, owner, name):
        self._name = name

    def __setattr__(self, name, value):
        '''Set any additional variable attribute into the default value.'''
        if name in self.__slots__ or name in self.__mutable_props:
            super().__setattr__(name, value)
        else:
            setattr(self._default_value, name, value)

    def __getattr__(self, name):
        '''Attribute lookup into the variable's value.'''
        def_val = self.__getattribute__('_p_default_value')

        # NOTE: This if below is necessary to avoid breaking the deepcopy
        # of instances of this class. Without it, a deepcopy of instances of
        # this class can return an instance of _UndefinedType when def_val
        # is Undefined. This is because _UndefinedType implements a custom
        # __deepcopy__ method.
        if def_val is not Undefined:
            try:
                return getattr(def_val, name)
            except AttributeError:
                '''Raise the AttributeError below.'''

        var_name = self.__getattribute__('_name')
        raise AttributeError(
            f'variable {var_name!r} has no attribute {name!r}'
        ) from None

    def _check_is_defined(self):
        if not self.is_defined():
            raise ReframeSyntaxError(
                f'variable {self._name!r} is not assigned a value'
            )

    def __str__(self):
        self._check_is_defined()
        return str(self._default_value)

    def __repr__(self):
        import reframe
        if hasattr(reframe, '__build_docs__'):
            return str(self)

        try:
            name = self.name
        except AttributeError:
            name = '<undef>'

        if self.is_defined():
            value = self._default_value
        else:
            value = '<undef>'

        return f'TestVar(name={name!r}, value={value!r})'

    def __bytes__(self):
        self._check_is_defined()
        return bytes(self._default_value)

    def __format__(self, *args):
        self._check_is_defined()
        return format(self._default_value, *args)

    def __lt__(self, other):
        self._check_is_defined()
        return self._default_value < other

    def __le__(self, other):
        self._check_is_defined()
        return self._default_value <= other

    def __eq__(self, other):
        self._check_is_defined()
        return self._default_value == other

    def __ne__(self, other):
        self._check_is_defined()
        return self._default_value != other

    def __gt__(self, other):
        self._check_is_defined()
        return self._default_value > other

    def __ge__(self, other):
        self._check_is_defined()
        return self._default_value >= other

    def __hash__(self):
        self._check_is_defined()
        return hash(self._default_value)

    def __bool__(self):
        self._check_is_defined()
        return bool(self._default_value)

    # Container Operators

    def __len__(self):
        self._check_is_defined()
        return len(self._default_value)

    def __len_hint__(self):
        self._check_is_defined()
        return self._default_value.__len_hint__()

    def __getitem__(self, key):
        self._check_is_defined()
        return self._default_value.__getitem__(key)

    def __setitem__(self, key, value):
        self._check_is_defined()
        self._default_value.__setitem__(key, value)

    def __delitem__(self, key):
        self._check_is_defined()
        self._default_value.__delitem__(key)

    def __missing__(self, key):
        self._check_is_defined()
        return self._default_value.__missing__(key)

    def __iter__(self):
        self._check_is_defined()
        return iter(self._default_value)

    def __reversed__(self):
        self._check_is_defined()
        return reversed(self._default_value)

    def __contains__(self, item):
        self._check_is_defined()
        return item in self._default_value

    # Numeric operators

    def __add__(self, other):
        self._check_is_defined()
        return self._default_value + other

    def __sub__(self, other):
        self._check_is_defined()
        return self._default_value - other

    def __mul__(self, other):
        self._check_is_defined()
        return self._default_value * other

    def __matmul__(self, other):
        self._check_is_defined()
        return self._default_value @ other

    def __truediv__(self, other):
        self._check_is_defined()
        return self._default_value / other

    def __floordiv__(self, other):
        self._check_is_defined()
        return self._default_value // other

    def __mod__(self, other):
        self._check_is_defined()
        return self._default_value % other

    def __divmod__(self, other):
        self._check_is_defined()
        return divmod(self._default_value, other)

    def __pow__(self, other, *modulo):
        self._check_is_defined()
        return pow(self._default_value, other, *modulo)

    def __lshift__(self, other):
        self._check_is_defined()
        return self._default_value << other

    def __rshift__(self, other):
        self._check_is_defined()
        return self._default_value >> other

    def __and__(self, other):
        self._check_is_defined()
        return self._default_value & other

    def __or__(self, other):
        self._check_is_defined()
        return self._default_value | other

    def __xor__(self, other):
        self._check_is_defined()
        return self._default_value ^ other

    def __radd__(self, other):
        self._check_is_defined()
        return other + self._default_value

    def __rsub__(self, other):
        self._check_is_defined()
        return other - self._default_value

    def __rmul__(self, other):
        self._check_is_defined()
        return other * self._default_value

    def __rmatmul__(self, other):
        self._check_is_defined()
        return other @ self._default_value

    def __rtruediv__(self, other):
        self._check_is_defined()
        return other / self._default_value

    def __rfloordiv__(self, other):
        self._check_is_defined()
        return other // self._default_value

    def __rmod__(self, other):
        self._check_is_defined()
        return other % self._default_value

    def __rdivmod__(self, other):
        self._check_is_defined()
        return divmod(other, self._default_value)

    def __rpow__(self, other, *modulo):
        self._check_is_defined()
        return pow(other, self._default_value, *modulo)

    def __rlshift__(self, other):
        self._check_is_defined()
        return other << self._default_value

    def __rrshift__(self, other):
        self._check_is_defined()
        return other >> self._default_value

    def __rand__(self, other):
        self._check_is_defined()
        return other & self._default_value

    def __ror__(self, other):
        self._check_is_defined()
        return other | self._default_value

    def __rxor__(self, other):
        self._check_is_defined()
        return other ^ self._default_value

    def __iadd__(self, other):
        self._check_is_defined()
        self._default_value += other
        return self._default_value

    def __isub__(self, other):
        self._check_is_defined()
        self._default_value -= other
        return self._default_value

    def __imul__(self, other):
        self._check_is_defined()
        self._default_value *= other
        return self._default_value

    def __imatmul__(self, other):
        self._check_is_defined()
        self._default_value @= other
        return self._default_value

    def __itruediv__(self, other):
        self._check_is_defined()
        self._default_value /= other
        return self._default_value

    def __ifloordiv__(self, other):
        self._check_is_defined()
        self._default_value //= other
        return self._default_value

    def __imod__(self, other):
        self._check_is_defined()
        self._default_value %= other
        return self._default_value

    def __ipow__(self, other, *modulo):
        self._check_is_defined()
        self._default_value = pow(self._default_value, other, *modulo)
        return self._default_value

    def __ilshift__(self, other):
        self._check_is_defined()
        self._default_value <<= other
        return self._default_value

    def __irshift__(self, other):
        self._check_is_defined()
        self._default_value >>= other
        return self._default_value

    def __iand__(self, other):
        self._check_is_defined()
        self._default_value &= other
        return self._default_value

    def __ior__(self, other):
        self._check_is_defined()
        self._default_value |= other
        return self._default_value

    def __ixor__(self, other):
        self._check_is_defined()
        self._default_value ^= other
        return self._default_value

    def __neg__(self):
        self._check_is_defined()
        return -self._default_value

    def __pos__(self):
        self._check_is_defined()
        return +self._default_value

    def __abs__(self):
        self._check_is_defined()
        return abs(self._default_value)

    def __invert__(self):
        self._check_is_defined()
        return ~self._default_value

    def __int__(self):
        self._check_is_defined()
        return int(self._default_value)

    def __float__(self):
        self._check_is_defined()
        return float(self._default_value)

    def __complex__(self):
        self._check_is_defined()
        return complex(self._default_value)

    def __round__(self, *ndigits):
        self._check_is_defined()
        return round(self._default_value, *ndigits)

    def __trunc__(self):
        self._check_is_defined()
        return math.trunc(self._default_value)

    def __floor__(self):
        self._check_is_defined()
        return math.floor(self._default_value)

    def __ceil__(self):
        self._check_is_defined()
        return math.ceil(self._default_value)


class ShadowVar(TestVar):
    '''A shadow instance of another variable.

    This is essentially a fully-fledged shallow copy of another variable. It
    is used during the construction of the class namespace to bring in scope a
    requested variable that is defined in a base class (see
    `MetaNamespace.__getitem__()`)

    We could not simply create a reference of the original variable in the
    current namespace, because we need a mechanism to differentiate the
    lowered variable from any redefinition, which is illegal.

    Also, we don't need a deep copy, since the shadow variable will replace
    the original variable in the newly constructed `VarSpace`.

    '''

    def __init__(self, other):
        for name in self.__slots__:
            setattr(self, name, getattr(other, name))

        self._warn_deprecation(DEPRECATE_RD)

    def __repr__(self):
        return super().__repr__().replace('TestVar', 'ShadowVar')


class VarSpace(namespaces.Namespace):
    '''Variable space of a regression test.

    A target class can be provided to the
    :func:`__init__` method, which is the regression test where the
    VarSpace is to be built. During this call to
    :func:`__init__`, the VarSpace inherits all the VarSpace from the base
    classes of the target class. After this, the VarSpace is extended with
    the information from the local variable space. If no target class is
    provided, the VarSpace is simply initialized as empty.
    '''

    def __init__(self, target_cls=None, illegal_names=None):
        # Set to register the variables already injected in the class
        self._injected_vars = set()
        super().__init__(target_cls, illegal_names,
                         ns_name='_rfm_var_space',
                         ns_local_name='_rfm_local_var_space')

    def join(self, other, cls):
        '''Join an existing VarSpace into the current one.

        :param other: instance of the VarSpace class.
        :param cls: the target class.
        '''
        for key, var in other.items():
            # Make doubly declared vars illegal. Note that this will be
            # triggered when inheriting from multiple RegressionTest classes.
            if key in self.vars:
                raise ReframeSyntaxError(
                    f'variable {key!r} is declared in more than one of the '
                    f'parent classes of class {cls.__qualname__!r}'
                )

            self.vars[key] = copy.deepcopy(var)

        # Inherited variables are copied in the current namespace, so we need
        # to update any aliases to point to the current namespace copies
        for var in self.vars.values():
            if var.is_alias():
                var.reset_target(self.vars[var.target.name])

        # Carry over the set of injected variables
        self._injected_vars.update(other._injected_vars)

    def extend(self, cls):
        '''Extend the VarSpace with the content in the LocalVarSpace.

        Merge the VarSpace inherited from the base classes with the
        LocalVarSpace. Note that the LocalVarSpace can also contain
        define and undefine actions on existing vars. Thus, since it
        does not make sense to define and undefine a var in the same
        class, the order on which the define and undefine functions
        are called is not preserved. In fact, applying more than one
        of these actions on the same var for the same local var space
        is disallowed.
        '''
        local_varspace = getattr(cls, self.local_namespace_name, False)
        for key, var in local_varspace.items():
            if isinstance(var, TestVar):
                # Disable redeclaring a variable
                if key in self.vars:
                    raise ReframeSyntaxError(
                        f'cannot redeclare the variable {key!r}'
                    )

                # Add a new var
                self.vars[key] = var

        local_varspace.clear()

        # If any previously declared variable was defined in the class body
        # by directly assigning it a value, retrieve this value from the class
        # namespace and update it into the variable space.
        _assigned_vars = set()
        for key, value in cls.__dict__.items():
            if key in self.vars:
                if isinstance(value, ShadowVar):
                    self.vars[key] = value
                else:
                    self.vars[key].define(value)

                _assigned_vars.add(key)
            elif value is Undefined:
                # Cannot be set as Undefined if not a variable
                raise ReframeSyntaxError(
                    f'{key!r} has not been declared as a variable'
                )

        # Delete the vars from the class __dict__.
        for key in _assigned_vars:
            delattr(cls, key)

    def sanity(self, cls, illegal_names):
        '''Sanity checks post-creation of the var namespace.

        By default, we make illegal to have any item in the namespace
        that clashes with a member of the target class unless this member
        was injected by this namespace.
        '''
        if illegal_names is None:
            illegal_names = set(dir(cls))

        for key in self._namespace:
            if key in illegal_names and key not in self._injected_vars:
                raise ReframeSyntaxError(
                    f'{key!r} already defined in class '
                    f'{cls.__qualname__!r}'
                )

    def inject(self, obj, cls):
        '''Insert the vars in the regression test.

        :param obj: The test object.
        :param cls: The test class.
        '''

        # Attribute injection is a special operation; the actual attribute
        # descriptor fields will be created and they will be assigned their
        # value; deprecations have been checked already during the class
        # construction, so we don't want to trigger them also here.
        with suppress_deprecations():
            self._inject(obj, cls)

    def _inject(self, obj, cls):
        for name, var in self.items():
            setattr(cls, name, var.field)
            getattr(cls, name).__set_name__(obj, name)

            # If the var is defined, set its value
            if var.is_defined():
                setattr(obj, name, var.default_value)

            # Track the variables that have been injected.
            self._injected_vars.add(name)

    @property
    def vars(self):
        return self._namespace

    @property
    def injected_vars(self):
        return self._injected_vars



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

#
# Functionality to use fixtures in ReFrame tests.
#

import sys
import copy
import itertools
import traceback
from collections.abc import Iterable, Mapping
from hashlib import sha256

import reframe.core.namespaces as namespaces
import reframe.core.runtime as runtime
import reframe.utility.udeps as udeps
import reframe.utility as utils
from reframe.core.exceptions import ReframeSyntaxError, what
from reframe.core.logging import getlogger


class FixtureData:
    '''Store raw data related to a fixture instance.

    The stored data is the fixture class variant number, a list with the valid
    environments, another list with the valid partitions, a dictionary with
    any fixture variables that may be set during the instantiation of the
    fixture class, and the scope used for the fixture.

    This data is required to instantiate the fixture.
    '''

    __slots__ = ('__data',)

    def __init__(self, variant, envs, parts, variables, scope, scope_enc):
        self.__data = (variant, envs, parts, variables, scope, scope_enc)

    @property
    def data(self):
        return self.__data

    @property
    def variant_num(self):
        return self.__data[0]

    @property
    def environments(self):
        return self.__data[1]

    @property
    def partitions(self):
        return self.__data[2]

    @property
    def variables(self):
        return self.__data[3]

    @property
    def scope(self):
        return self.__data[4]

    @property
    def scope_enc(self):
        return self.__data[5]

    def mashup(self):
        s = f'{self.variant_num}/{self.scope_enc}'
        if self.variables:
            s += '/' + '&'.join(f'{k}={self.variables[k]}'
                                for k in sorted(self.variables))

        return sha256(s.encode('utf-8')).hexdigest()[:8]


class FixtureRegistry:
    '''Regression test fixture registry.

    This registry consists of a dictionary where the keys are the fixture
    classes and the values are dictionaries containing raw data that
    identifies each of the registered fixtures to be instantiated from the
    fixture class.

    This registry mangles the fixture name to account for the fixture scope and
    other fixture-specific details that make the fixture unique. This name
    mangling ensures that two different fixtures from the same class do not
    have the same name. In this context, a fixture variant is determined by the
    fixture scope and a 4-element tuple containing the class variant number, a
    list with the valid environments, another list with the valid partitions
    and a key-value pair mapping with any fixture variables that may be set
    during the instantiation of the fixture class.

    Fixtures are added to the registry with the :func:`add` method, and the
    registered fixtures can be instantiated with the :func:`instantiate_all`
    method.

    Since the fixture name is mangled to be unique, a fixture that
    modifies its :attr:`name` attribute may result in an undefined behaviour.

    :meta private:
    '''

    def __init__(self):
        self._registry = dict()

        # Build an index map to access the system partitions
        sys_part = runtime.runtime().system.partitions
        self._env_by_part = {
            p.fullname: {e.name for e in p.environs} for p in sys_part
        }

        # Store the system name for name-mangling purposes
        self._sys_name = runtime.runtime().system.name

    def add(self, fixture, variant_num, parent_test):
        '''Register a fixture.

        This method mangles the fixture name, ensuring that different fixture
        variants (see above for the definition of a fixture variant in the
        context of a fixture registry) have a different name. This method
        is aware of the public members from the :class:`TestFixture`.
        Fixtures `steal` the valid environments and valid partitions from
        their parent test, where the number of combinations that trickle
        down into the fixture varies depending on the fixture's scope.
        The rationale is as follows for the different scopes:
         - session: Only one environment+partition combination per fixture.
               This kind of fixture may be shared across all tests. The name
               for fixtures with this scope is mangled with the system name.
               This is necesssary to avoid name conflicts with classes that
               are used both as regular tests and fixtures with session scope.
         - partition: Only one environment per partition. This kind of fixture
               may be shared amongst all the tests running on the same
               partition. The name is mangled to include the partition where
               fixture will run.
         - environment: One fixture per available environment+partition
               combination. This kind of fixture may be shared only with
               other tests that execute on the same environment+partition
               combination. The name is mangled to contain the partition and
               the environment where the fixture will run.
         - test: Use the environments and partitions from the parent test
               without any modifications. Fixtures using this scope are
               private to the parent test and they will not be shared with
               any other test in the session. The fixture name is mangled to
               contain the name of the parent test the fixture belongs to.

        If a fixture modifies the default value of any of its attributes,
        these modified variables are always mangled into the fixture name
        regardless of the scope used. These variables are sorted by their
        name before they're accounted for into the name mangling, so the
        order on which they're specified is irrelevant.

        This method returns a list with the mangled names of the newly
        registered fixtures.

        :param fixture: An instance of :class:`TestFixture`.
        :param variant_num: The variant index for the given ``fixture``.
        :param parent_test: The parent test.
        '''

        cls = fixture.cls
        scope = fixture.scope
        fname = fixture.cls.variant_name(variant_num)
        variables = fixture.variables
        reg_names = []
        self._registry.setdefault(cls, dict())

        # Mangle the fixture name with the modified variables (sorted by keys)
        if variables:
            vname = ''.join(
                (f'%{k}={utils.toalphanum(str(v))}' for k, v
                 in sorted(variables.items()))
            )
            fname += '_' + sha256(vname.encode('utf-8')).hexdigest()[:8]

        # Select only the valid partitions
        try:
            valid_sysenv = runtime.valid_sysenv_comb(
                parent_test.valid_systems,
                parent_test.valid_prog_environs
            )
        except AttributeError as e:
            msg = e.args[0] + f' in test {parent_test.display_name!r}'
            raise ReframeSyntaxError(msg) from None

        # Return if there are no valid system/environment combinations
        if not valid_sysenv:
            return []

        # Register the fixture
        if scope == 'session':
            # The name is mangled with the system name
            # Pick the first valid system/environment combination
            pname, ename = None, None
            for part, environs in valid_sysenv.items():
                pname = part.fullname
                for env in environs:
                    ename = env.name
                    break

            if ename is None:
                # No valid environments found
                return []

            # Register the fixture
            fixt_data = FixtureData(variant_num, [ename], [pname],
                                    variables, scope, self._sys_name)
            name = f'{cls.__name__}_{fixt_data.mashup()}'
            self._registry[cls][name] = fixt_data
            reg_names.append(name)
        elif scope == 'partition':
            for part, environs in valid_sysenv.items():
                # The mangled name contains the full partition name
                # Select an environment supported by the partition
                pname = part.fullname
                try:
                    ename = environs[0].name
                except IndexError:
                    continue

                # Register the fixture
                fixt_data = FixtureData(variant_num, [ename], [pname],
                                        variables, scope, pname)
                name = f'{cls.__name__}_{fixt_data.mashup()}'
                self._registry[cls][name] = fixt_data
                reg_names.append(name)
        elif scope == 'environment':
            for part, environs in valid_sysenv.items():
                for env in environs:
                    # The mangled name contains the full part and env names
                    # Register the fixture
                    pname, ename = part.fullname, env.name
                    fixt_data = FixtureData(variant_num, [ename], [pname],
                                            variables, scope,
                                            f'{pname}+{ename}')
                    name = f'{cls.__name__}_{fixt_data.mashup()}'
                    self._registry[cls][name] = fixt_data
                    reg_names.append(name)
        elif scope == 'test':
            # The mangled name contains the parent test name.

            # Register the fixture
            fixt_data = FixtureData(variant_num,
                                    list(parent_test.valid_prog_environs),
                                    list(parent_test.valid_systems),
                                    variables, scope, parent_test.unique_name)
            name = f'{cls.__name__}_{fixt_data.mashup()}'
            self._registry[cls][name] = fixt_data
            reg_names.append(name)

        return reg_names

    def update(self, other):
        '''Extend the current registry with the items from another registry.

        In the event of a clash, the elements from ``other`` take precedence.
        Clashes are allowed because they would only happen when two fixtures
        are equivalent (for example, for a fixture with ``'session'`` scope is
        irrelevant which partition or environment is selected to run the
        fixture).
        '''

        self._is_registry(other)
        for cls, variants in other._registry.items():
            self._registry.setdefault(cls, dict())
            for name, args in variants.items():
                self._registry[cls][name] = args

    def difference(self, other):
        '''Build a new registry taking the difference with another registry

        The resulting registry contains the elements from the current registry
        that are not present in ``other``.
        '''

        self._is_registry(other)
        ret = FixtureRegistry()
        for cls, variants in self._registry.items():
            if cls in other:
                other_variants = other._registry[cls]
                for name, args in variants.items():
                    if name not in other_variants:
                        ret._registry.setdefault(cls, dict())
                        ret._registry[cls][name] = args
            else:
                ret._registry[cls] = copy.deepcopy(variants)

        return ret

    def instantiate_all(self):
        '''Instantiate all the fixtures in the registry.'''

        ret = []
        for cls, variants in self._registry.items():
            for name, args in variants.items():
                varnum, penv, part, variables, *_ = args.data

                # Set the fixture name and stolen env and part from the parent,
                # alongside the other variables specified during the fixture's
                # declaration.
                fixtvars = {
                    'valid_prog_environs': penv,
                    'valid_systems': part,
                    **variables
                }

                try:
                    # Instantiate the fixture
                    inst = cls(variant_num=varnum, fixt_name=name,
                               fixt_data=args, fixt_vars=fixtvars)
                except Exception:
                    exc_info = sys.exc_info()
                    getlogger().warning(
                        f"skipping fixture {name!r}: "
                        f"{what(*exc_info)} "
                        f"(rerun with '-v' for more information)"
                    )
                    getlogger().verbose(traceback.format_exc())
                else:
                    ret.append(inst)

        return ret

    def _filter_valid_partitions(self, candidate_parts):
        return [p for p in candidate_parts if p in self._env_by_part]

    def _filter_valid_environs(self, part, candidate_environs):
        return [e for e in cadidate_environs if e in self._env_by_part[part]]

    def _is_registry(self, other):
        if not isinstance(other, FixtureRegistry):
            raise TypeError('other is not a FixtureRegistry')

    def __getitem__(self, cls):
        '''Return the sub-dictionary for a given fixture class.

        The keys are the mangled fixture names and the values are the raw data
        to instantiate the fixture class with.
        '''
        return self._registry.get(cls, dict())

    def __contains__(self, cls):
        return cls in self._registry


class TestFixture:
    '''Insert a new fixture in the current test.

    A fixture is a regression test that creates, prepares and/or manages a
    resource for another regression test. Fixtures may contain other fixtures
    and so on, forming a directed acyclic graph. A parent fixture (or a
    regular regression test) requires the resources managed by its child
    fixtures in order to run, and it may only access these fixture resources
    after its ``setup`` pipeline stage. The execution of parent fixtures is
    postponed until all their respective children have completed execution.
    However, the destruction of the resources managed by a fixture occurs in
    reverse order, only after all the parent fixtures have been destroyed.
    This destruction of resources takes place during the ``cleanup`` pipeline
    stage of the regression test. Fixtures must not define the members
    :attr:`~reframe.core.pipeline.RegressionTest.valid_systems` and
    :attr:`~reframe.core.pipeline.RegressionTest.valid_prog_environs`. These
    variables are defined based on the values specified in the parent test,
    ensuring that the fixture runs with a suitable system partition and
    programming environment combination. A fixture's
    :attr:`~reframe.core.pipeline.RegressionTest.name` attribute may be
    internally mangled depending on the arguments passed during the fixture
    declaration. Hence, manually setting or modifying the
    :attr:`~reframe.core.pipeline.RegressionTest.name` attribute in the
    fixture class is disallowed, and breaking this restriction will result in
    undefined behavior.

    .. warning::
       The fixture name mangling is considered an internal framework mechanism
       and it may change in future versions without any notice. Users must not
       express any logic in their tests that relies on a given fixture name
       mangling scheme.


    By default, the resources managed by a fixture are private to the parent
    test. However, it is possible to share these resources across different
    tests by passing the appropriate fixture ``scope`` argument. The different
    scope levels are independent from each other and a fixture only executes
    once per scope, where all the tests that belong to that same scope may use
    the same resources managed by a given fixture instance. The available
    scopes are:

      - **session**: This scope encloses all the tests and fixtures that run
        in the full ReFrame session. This may include tests that use different
        system partition and programming environment combinations. The fixture
        class must derive from
        :class:`~reframe.core.pipeline.RunOnlyRegressionTest` to avoid any
        implicit dependencies on the partition or the programming environment
        used.

      - **partition**: This scope spans across a single system partition. This
        may include different tests that run on the same partition but use
        different programming environments. Fixtures with this scope must be
        independent of the programming environment, which restricts the
        fixture class to derive from
        :class:`~reframe.core.pipeline.RunOnlyRegressionTest`.

      - **environment**: The extent of this scope covers a single combination
        of system partition and programming environment. Since the fixture is
        guaranteed to have the same partition and programming environment as
        the parent test, the fixture class can be any derived class from
        :class:`~reframe.core.pipeline.RegressionTest`. * **test**: This scope
        covers a single instance of the parent test, where the resources
        provided by the fixture are exclusive to each parent test instance.
        The fixture class can be any derived class from
        :class:`~reframe.core.pipeline.RegressionTest`.

    Rather than specifying the scope at the fixture class definition, ReFrame
    fixtures set the scope level from the consumer side (i.e. when used by
    another test or fixture). A test may declare multiple fixtures using the
    same class, where fixtures with different scopes are guaranteed to point
    to different instances of the fixture class. On the other hand, when two
    or more fixtures use the same fixture class and have the same scope, these
    different fixtures will point to the same underlying resource if the
    fixtures refer to the same :ref:`variant<test-variants>` of the fixture
    class. The example below illustrates the different fixture scope usages:

    .. code:: python

       class MyFixture(rfm.RunOnlyRegressionTest):
          my_var = variable(int, value=1)
          ...


       @rfm.simple_test
       class TestA(rfm.RegressionTest):
           valid_systems = ['p1', 'p2']
           valid_prog_environs = ['e1', 'e2']

           # Fixture shared throughout the full session
           f1 = fixture(MyFixture, scope='session')

           # Fixture shared for each supported partition
           f2 = fixture(MyFixture, scope='partition')

           # Fixture shared for each supported part+environ
           f3 = fixture(MyFixture, scope='environment')

           # Fixture private evaluation of MyFixture
           f4 = fixture(MyFixture, scope='test')
           ...


       @rfm.simple_test
       class TestB(rfm.RegressionTest):
           valid_systems = ['p1']
           valid_prog_environs = ['e1']

           # Another private instance of MyFixture
           f1 = fixture(MyFixture, scope='test')

           # Same as f3 in TestA for p1 + e1
           f2 = fixture(MyFixture, scope='environment')

           # Same as f1 in TestA
           f3 = fixture(MyFixture, scope='session')
           ...

           @run_after('setup')
           def access_fixture_resources(self):
               # Dummy pipeline hook to illustrate fixture resource access
               assert self.f1.my_var is not self.f2.my_var
               assert self.f1.my_var is not self.f3.my_var


    :class:`TestA` supports two different valid systems and another two valid
    programming environments. Assuming that both environments are supported by
    each of the system partitions ``'p1'`` and ``'p2'``, this test will
    execute a total of four times. This test uses the very simple
    :class:`MyFixture` fixture multiple times using different scopes, where
    fixture ``f1`` (session scope) will be shared across the four test
    instances, and fixture ``f4`` (test scope) will be executed once per test
    instance. On the other hand, ``f2`` (partition scope) will run once per
    partition supported by test :class:`TestA`, and the multiple per-partition
    executions (i.e. for each programming environment) will share the same
    underlying resource for ``f2``. Lastly, ``f3`` will run a total of four
    times, which is once per partition and environment combination. This
    simple :class:`TestA` shows how multiple instances from the same test can
    share resources, but the real power behind fixtures is illustrated with
    :class:`TestB`, where this resource sharing is extended across different
    tests. For simplicity, :class:`TestB` only supports a single partition
    ``'p1'`` and programming environment ``'e1'``, and similarly to
    :class:`TestA`, ``f1`` (test scope) causes a private evaluation of the
    fixture :class:`MyFixture`. However, the resources managed by fixtures
    ``f2`` (environment scope) and ``f3`` (session scope) are shared with
    :class:`Test1`.

    Fixtures are treated by ReFrame as first-class ReFrame tests, which means
    that these classes can use the same built-in functionalities as in regular
    tests decorated with
    :func:`@rfm.simple_test<reframe.core.decorators.simple_test>`. This
    includes the :func:`~reframe.core.pipeline.RegressionMixin.parameter`
    built-in, where fixtures may have more than one
    :ref:`variant<test-variants>`. When this occurs, a parent test may select
    to either treat a parameterized fixture as a test parameter, or instead,
    to gather all the fixture variants from a single instance of the parent
    test. In essence, fixtures implement `fork-join` model whose behavior may
    be controlled through the ``action`` argument. This argument may be set to
    one of the following options:

      - **fork**: This option parameterizes the parent test as a function of
        the fixture variants. The fixture handle will resolve to a single
        instance of the fixture.

      - **join**: This option gathers all the variants from a fixture into a
        single instance of the parent test. The fixture handle will point to a
        list containing all the fixture variants.

    A test may declare multiple fixtures with different ``action`` options,
    where the default ``action`` option is ``'fork'``. The example below
    illustrates the behavior of these two different options.

    .. code:: python

       class ParamFix(rfm.RegressionTest):
           p = parameter(range(5)) # A simple test parameter
           ...


       @rfm.simple_test
       class TestC(rfm.RegressionTest):
           # Parameterize TestC for each ParamFix variant
           f = fixture(ParamFix, action='fork')
           ...

           @run_after('setup')
           def access_fixture_resources(self):
               print(self.f.p) # Prints the fixture's variant parameter value


       @rfm.simple_test
       class TestD(rfm.RegressionTest):
           # Gather all fixture variants into a single test
           f = fixture(ParamFix, action='join')
           ...

           @run_after('setup')
           def reduce_range(self):
               # Sum all the values of p for each fixture variant
               res = functools.reduce(lambda x, y: x+y,
                                      (fix.p for fix in self.f))
               n = len(self.f)-1
               assert res == (n*n + n)/2

    Here :class:`ParamFix` is a simple fixture class with a single parameter.
    When the test :class:`TestC` uses this fixture with a ``'fork'`` action,
    the test is implicitly parameterized over each variant of
    :class:`ParamFix`. Hence, when the :func:`access_fixture_resources`
    post-setup hook accesses the fixture ``f``, it only access a single
    instance of the :class:`ParamFix` fixture. On the other hand, when this
    same fixture is used with a ``'join'`` action by :class:`TestD`, the test
    is not parameterized and all the :class:`ParamFix` instances are gathered
    into ``f`` as a list. Thus, the post-setup pipeline hook
    :func:`reduce_range` can access all the fixture variants and compute a
    reduction of the different ``p`` values.

    When declaring a fixture, a parent test may select a subset of the fixture
    variants through the ``variants`` argument. This variant selection can be
    done by either passing an iterable containing valid variant indices (see
    :ref:`test-variants` for further information on how the test variants are
    indexed), or instead, passing a mapping with the parameter name (of the
    fixture class) as keys and filtering functions as values. These filtering
    functions are unary functions that return the value of a boolean
    expression on the values of the specified parameter, and they all must
    evaluate to :class:`True` for at least one of the fixture class variants.
    See the example below for an illustration on how to filter-out fixture
    variants.

    .. code:: python

       class ComplexFixture(rfm.RegressionTest):
           # A fixture with 400 different variants.
           p0 = parameter(range(100))
           p1 = parameter(['a', 'b', 'c', 'd'])
           ...

       @rfm.simple_test
       class TestE(rfm.RegressionTest):
           # Select the fixture variants with boolean conditions
           foo = fixture(ComplexFixture,
                         variants={'p0': lambda x: x<10,
                                   'p1': lambda x: x=='d'})

           # Select the fixture variants by index
           bar = fixture(ComplexFixture, variants=range(300,310))
           ...

    A parent test may also specify the value of different variables in the
    fixture class to be set before its instantiation. Each variable must have
    been declared in the fixture class with the
    :func:`~reframe.core.pipeline.RegressionMixin.variable` built-in,
    otherwise it is silently ignored. This variable specification is
    equivalent to deriving a new class from the fixture class, and setting
    these variable values in the class body of a newly derived class.
    Therefore, when fixture declarations use the same fixture class and pass
    different values to the ``variables`` argument, the fixture class is
    interpreted as a different class for each of these fixture declarations.
    See the example below.

    .. code:: python

       class Fixture(rfm.RegressionTest):
           v = variable(int, value=1)
           ...

       @rfm.simple_test
       class TestF(rfm.RegressionTest):
           foo = fixture(Fixture)
           bar = fixture(Fixture, variables={'v':5})
           baz = fixture(Fixture, variables={'v':10})
           ...

           @run_after('setup')
           def print_fixture_variables(self):
               print(self.foo.v) # Prints 1
               print(self.bar.v) # Prints 5
               print(self.baz.v) # Prints 10

    The test :class:`TestF` declares the fixtures ``foo``, ``bar`` and ``baz``
    using the same :class:`Fixture` class. If no variables were set in ``bar``
    and ``baz``, this would result into the same fixture being declared
    multiple times in the same scope (implicitly set to ``'test'``), which
    would lead to a single instance of :class:`Fixture` being referred to by
    ``foo``, ``bar`` and ``baz``. However, in this case ReFrame identifies
    that the declared fixtures pass different values to the ``variables``
    argument in the fixture declaration, and executes these three fixtures
    separately.

    .. note::
       Mappings passed to the ``variables`` argument that define the same
       class variables in different order are interpreted as the same value.
       The two fixture declarations below are equivalent, and both ``foo`` and
       ``bar`` will point to the same instance of the fixture class
       :class:`MyResource`.

       .. code:: python

         foo = fixture(MyResource, variables={'a':1, 'b':2})
         bar = fixture(MyResource, variables={'b':2, 'a':1})


    **Early access to fixture objects**

    The test instance represented by a fixture can be accessed fully from
    within a test only after the setup stage. The reason for that is that
    fixtures eventually translate into test dependencies and access to the
    parent dependencies cannot happen before the this stage.

    However, it is often useful, especially in the case of parameterized
    fixtures, to be able to access the fixture parameters earlier, e.g., in a
    post-init hook in order to properly set the
    :attr:`~reframe.core.pipeline.RegressionTest.valid_systems` and
    :attr:`~reframe.core.pipeline.RegressionTest.valid_prog_environs` of the
    test. These attributes cannot be set later than the test's initialization
    in order to have an effect.

    For this reason, early access to fixture objects is allowed *only* for
    retrieving their parameters.

    .. code-block:: python

       class Fixture(rfm.RegressionTest):
           x = parameter([1, 2, 3])


       class Test(rfm.RunOnlyRegressionTest):
           foo = fixture(Fixture)
           executable = './myexec'
           valid_prog_environs = ['*']

           @run_after('init')
           def early_access(self):
                # Only fixture parameters can be accessed here!
                if self.foo.x == 1:
                    self.valid_systems = ['sys1]
                else:
                    self.valid_systems = ['sys2']

           @run_after('setup')
           def normal_access(self):
               # Any test attribute of the associated fixture test can be
               # accessed here
               self.executable_opts = [
                   '-i', os.path.join(self.foo.stagedir, 'input.txt')
               ]


    During test initialization, ReFrame binds the :attr:`foo` name to a proxy
    object that holds the parameterization of the target fixture. This proxy
    object is recursive, so that if fixture :attr:`foo` contained another
    fixture named :attr:`bar`, it would allow you to access any parameters of
    that fixture with ``self.foo.bar.param``.

    During the test setup stage, the :attr:`foo`'s binding changes and it is
    now bound to the exact test instance that was executed for the target test
    instance.

    :param cls: A class derived from
        :class:`~reframe.core.pipeline.RegressionTest` that manages a given
        resource. The base from this class may be further restricted to other
        derived classes of :class:`~reframe.core.pipeline.RegressionTest`
        depending on the ``scope`` parameter.

    :param scope: Sets the extent to which other regression tests may share
        the resources managed by a fixture. The available scopes are, from
        more to less restrictive, ``'test'``, ``'environment'``,
        ``'partition'`` and ``'session'``. By default a fixture's scope is set
        to ``'test'``, which makes the resource private to the test that uses
        the fixture. This means that when multiple regression tests use the
        same fixture class with a ``'test'`` scope, the fixture will run once
        per regression test. When the scope is set to ``'environment'``, the
        resources managed by the fixture are shared across all the tests that
        use the fixture and run on the same system partition and use the same
        programming environment. When the scope is set to ``'partition'``, the
        resources managed by the fixture are shared instead across all the
        tests that use the fixture and run on the same system partition.
        Lastly, when the scope is set to ``'session'``, the resources managed
        by the fixture are shared across the full ReFrame session. Fixtures
        with either ``'partition'`` or ``'session'`` scopes may be shared
        across different regression tests under different programming
        environments, and for this reason, when using these two scopes, the
        fixture class ``cls`` is required to derive from
        :class:`~reframe.core.pipeline.RunOnlyRegressionTest`.

    :param action: Set the behavior of a parameterized fixture to either
        ``'fork'`` or ``'join'``. With a ``'fork'`` action, a parameterized
        fixture effectively parameterizes the regression test. On the other
        hand, a ``'join'`` action gathers all the fixture variants into the
        same instance of the regression test. By default, the ``action``
        parameter is set to ``'fork'``.

    :param variants: Filter or sub-select a subset of the variants from a
        parameterized fixture. This argument can be either an iterable with
        the indices from the desired variants, or a mapping containing unary
        functions that return the value of a boolean expression on the values
        of a given parameter.

    :param variables: Mapping to set the values of fixture's variables. The
        variables are set after the fixture class has been created (i.e. after
        the class body has executed) and before the fixture class is
        instantiated.


    .. versionadded:: 3.9.0
    .. versionchanged:: 3.11.0
       Allow early access of fixture objects.

    '''

    def __init__(self, cls, *, scope='test', action='fork', variants='all',
                 variables=None):
        # Validate the fixture class: We can't use isinstance here because of
        # circular imports.
        rfm_kind = getattr(cls, '_rfm_regression_class_kind', 0)
        if rfm_kind == 0:
            raise ValueError(
                f"{cls.__qualname__!r} must be a derived class from "
                f"'RegressionTest'"
            )
        elif rfm_kind & 1:
            if scope in {'session', 'partition'}:
                raise ValueError(
                    f'incompatible scope for fixture {cls.__qualname__!r}; '
                    f'scope {scope!r} only supports run-only fixtures.'
                )

        # Check that the fixture class is not an abstract test.
        if cls.is_abstract():
            raise ValueError(
                f'class {cls.__qualname__!r} has undefined parameters'
            )

        # Validate the scope
        if scope not in ('session', 'partition', 'environment', 'test'):
            raise ValueError(
                f'invalid scope for fixture {cls.__qualname__!r}: {scope!r}'
            )

        # Validate the action
        if action not in ('fork', 'join'):
            raise ValueError(
                f'invalid action for fixture {cls.__qualname__!r}: {action!r}'
            )

        self._cls = cls
        self._scope = scope
        self._action = action
        if isinstance(variants, Mapping):
            # If the variants are passed as a mapping, this argument is
            # expected to contain conditions to filter the fixture's
            # parameter space.
            self._variants = tuple(cls.get_variant_nums(**variants))
        elif isinstance(variants, Iterable) and not isinstance(variants, str):
            self._variants = tuple(variants)
        elif variants == 'all':
            self._variants = tuple(range(cls.num_variants))
        else:
            raise ValueError(
                f'invalid variants specified for fixture {cls.__qualname__!r}'
            )

        # Check that we have some variants
        if len(self._variants) == 0:
            raise ValueError('fixture does not have any variants')

        if variables and not isinstance(variables, Mapping):
            raise TypeError(
                "the argument 'variables' must be a mapping."
            )
        elif variables is None:
            variables = {}

        # Store the variables dict
        self._variables = variables

    @property
    def cls(self):
        '''The underlying RegressionTest class.'''
        return self._cls

    @property
    def scope(self):
        '''The fixture scope.'''
        return self._scope

    @property
    def action(self):
        '''Action specified on this fixture.'''
        return self._action

    @property
    def variants(self):
        '''The specified variants for this fixture.'''
        return self._variants

    @property
    def fork_variants(self):
        '''Collection of fixture variant sets that the parent test will fork.

        This function returns a tuple, where each of the elements represents
        a single fork. These elements are iterables containing a set of the
        fixture variant indices that a single fork will reduce. For example,
        for a returned tuple ``((1, 2), (3, 4))``, the parent test using this
        fixture will fork twice, where the first fork will reduce the fixture
        variants #1 and #2, and the second fork will reduce the fixture
        variants #3 and #4.

        With a ``'fork'`` action, the returned tuple will have as many elements
        as specified fixture variants, where each of the elements is a
        one-element tuple containing a unique fixture variant ID (with the
        above example, this is ``((1,), (2,), (3,), (4,))``). With a ``'join'``
        action, the returned tuple will contain a single iterable containing
        all the specified fixture variants (``((1, 2, 3, 4),)``).

        .. note::
          This forking model could be made fully customisable to the user by
          simply allowing to pass a function as the ``action`` argument that
          process the variants in their custom way to generate the tuple of
          iterables. However, this would conceptually overlap with the
          ``variants`` argument, since a user could pass a function that does
          not use the specified fixture variants at all.
        '''
        if self._action == 'join':
            return tuple((self.variants,))
        else:
            return tuple((v,) for v in self.variants)

    @property
    def variables(self):
        '''Variables to be set in the test.'''
        return self._variables


class FixtureProxy:
    def __init__(self, fixture_info):
        for k, v in fixture_info['params'].items():
            setattr(self, k, v)

        for k, v in fixture_info['fixtures'].items():
            if not isinstance(v, tuple):
                setattr(self, k, FixtureProxy(v))


class FixtureSpace(namespaces.Namespace):
    '''Regression test fixture space.

    The fixture space is first built by joining the available fixture spaces
    in the base classes, and later extended by the locally defined fixtures
    that are expected in the local fixture space. Defining fixtures with the
    same name in more than one of the base classes is disallowed. However,
    a fixture defined in a base class can be overridden by a fixture defined
    in the derived class under the same name.

    The fixture injection occurs on an instance of the target class. The
    fixtures are first grouped in a fixture registry, which is then injected
    into the target instance under the ``_rfm_fixture_registry`` attribute.
    '''

    def __init__(self, target_cls=None, illegal_names=None):
        super().__init__(target_cls, illegal_names,
                         ns_name='_rfm_fixture_space',
                         ns_local_name='_rfm_local_fixture_space')

        # Store all fixture variant combinations to allow random access.
        self.__variant_combinations = tuple(
            itertools.product(
                *(list(f.fork_variants for f in self.fixtures.values()))
            )
        )

    def join(self, other, cls):
        '''Join other fixture spaces into the current one.

        :param other: instance of the FixtureSpace class.
        :param cls: the target class.
        '''
        for key, value in other.fixtures.items():
            if key in self.fixtures:
                raise ReframeSyntaxError(
                    f'fixture space conflict: '
                    f'fixture {key!r} is defined in more than '
                    f'one base class of {cls.__qualname__!r}'
                )

            self.fixtures[key] = value

    def extend(self, cls):
        '''Extend the inherited fixture space with the local fixture space.'''
        local_fixture_space = getattr(cls, self.local_namespace_name, False)
        while local_fixture_space:
            name, fixture = local_fixture_space.popitem()
            self.fixtures[name] = fixture

        # If any previously declared fixture was defined in the class body
        # by directly assigning it a value, raise an error. Fixtures must be
        # changed using the `x = fixture(...)` syntax.
        for key in self.fixtures:
            if key in cls.__dict__:
                raise ReframeSyntaxError(
                    f'fixture {key!r} can only be redefined through the '
                    f'fixture built-in'
                )

    def inject(self, obj, cls=None, fixtures_index=None):
        '''Build fixture registry and inject it in the parent's test instance.

        A fixture steals the valid_systems and valid_prog_environs from the
        parent tests, and these attributes could be set during the parent
        test's instantiation. Similarly, the fixture registry requires of the
        parent test's full name to build unique IDs for fixtures with the
        ``'test'`` scope (i.e. fixtures private to a parent test).

        :param obj: Parent test's instance.
        :param cls: Parent test's class.
        :param fixtures_index: Index representing a point in the fixture
            space. This point represents a unique variant combination of all
            the fixtures present in the fixture space.

        .. note::
           This function is aware of the implementation of the
           :class:`reframe.core.pipeline.RegressionTest` class.
        '''

        # Nothing to do if the fixture space is empty
        if not self.fixtures or fixtures_index is None:
            return

        # Get the unique fixture variant combiantion (as a k-v map) for
        # the given index.
        try:
            fixture_variants = self[fixtures_index]
        except IndexError:
            raise RuntimeError(
                f'fixture space index out of range for '
                f'{obj.__class__.__qualname__}'
            ) from None

        # Create the fixture registry
        obj._rfm_fixture_registry = FixtureRegistry()

        # Register the fixtures
        for name, fixture in self.fixtures.items():
            dep_names = []
            for variant in fixture_variants[name]:
                # Register all the variants and track the fixture names
                dep_names += obj._rfm_fixture_registry.add(fixture,
                                                           variant, obj)

            # Add dependencies
            if fixture.scope == 'session':
                dep_kind = udeps.fully
            elif fixture.scope == 'partition':
                dep_kind = udeps.by_part
            else:
                dep_kind = udeps.by_case

            # Inject the dependency
            for dep_name in dep_names:
                obj.depends_on(dep_name, dep_kind)

    def __iter__(self):
        '''Walk through all index combinations for all fixtures.'''
        yield from self.__variant_combinations

    def __len__(self):
        if not self.fixtures:
            return 1

        return len(self.__variant_combinations)

    def __getitem__(self, key):
        '''Access an element in the fixture space.

        If the key is an integer, this function will return a mapping with the
        variant numbers of each of the fixtures for the provided point in the
        fixture space. In this case, the fixture must be an index in the range
        of ``[0, len(self))``.
        If the key is just a fixture name, this function will return the
        underlying fixture object with that name.
        '''
        if isinstance(key, int):
            ret = {}
            f_ids = self.__variant_combinations[key]
            for i, f in enumerate(self.fixtures):
                ret[f] = f_ids[i]

            return ret

        return self.fixtures[key]

    @property
    def fixtures(self):
        return self._namespace



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import abc
import copy
import os
import signal
import sys
import time
import weakref

import reframe.core.logging as logging
import reframe.core.runtime as runtime
import reframe.frontend.dependencies as dependencies
import reframe.utility.jsonext as jsonext
from reframe.core.exceptions import (AbortTaskError,
                                     JobNotStartedError,
                                     FailureLimitError,
                                     ForceExitError,
                                     SkipTestError,
                                     TaskExit)
from reframe.core.schedulers.local import LocalJobScheduler
from reframe.frontend.printer import PrettyPrinter
from reframe.frontend.statistics import TestStats

ABORT_REASONS = (AssertionError, FailureLimitError,
                 KeyboardInterrupt, ForceExitError)


class TestCase:
    '''A combination of a regression check, a system partition
    and a programming environment.
    '''

    def __init__(self, check, partition, environ):
        self._check_orig = check
        self._check = check
        self._partition = partition
        self._environ = environ
        self._deps = []
        self._is_ready = False

        # Incoming dependencies
        self.in_degree = 0

        # Level in the dependency chain
        self.level = 0

    def __iter__(self):
        # Allow unpacking a test case with a single liner:
        #       c, p, e = case
        return iter([self._check, self._partition, self._environ])

    def __hash__(self):
        return (hash(self.check.unique_name) ^
                hash(self.partition.fullname) ^
                hash(self.environ.name))

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        return (self.check.unique_name == other.check.unique_name and
                self.environ.name == other.environ.name and
                self.partition.fullname == other.partition.fullname)

    def __repr__(self):
        c = self.check.unique_name if self.check else None
        p = self.partition.fullname  if self.partition else None
        e = self.environ.name if self.environ else None
        return f'({c!r}, {p!r}, {e!r})'

    def prepare(self):
        '''Prepare test case for sending down the test pipeline'''
        if self._is_ready:
            return

        self._check = copy.deepcopy(self._check)
        self._check._case = weakref.ref(self)
        self._is_ready = True

    @property
    def check(self):
        return self._check

    @property
    def partition(self):
        return self._partition

    @property
    def environ(self):
        return self._environ

    @property
    def deps(self):
        return self._deps

    @property
    def num_dependents(self):
        return self.in_degree

    def clone(self):
        # Return a fresh clone, i.e., one based on the original check
        return TestCase(self._check_orig, self._partition, self._environ)


@logging.time_function
def generate_testcases(checks, prepare=False):
    '''Generate concrete test cases from checks.

    If `prepare` is true then each of the cases will also be prepared for
    being sent to the test pipeline. Note that setting this to true may slow down
    the test case generation.

    '''

    rt = runtime.runtime()
    cases = []
    for c in checks:
        valid_comb = runtime.valid_sysenv_comb(c.valid_systems,
                                               c.valid_prog_environs)
        for part, environs in valid_comb.items():
            for env in environs:
                case = TestCase(c, part, env)
                if prepare:
                    case.prepare()

                cases.append(case)

    return cases


class RegressionTask:
    '''A class representing a :class:`RegressionTest` through the regression
    pipeline.'''

    def __init__(self, case, listeners=[]):
        self._case = case
        self._failed_stage = None
        self._current_stage = 'startup'
        self._exc_info = (None, None, None)
        self._listeners = list(listeners)
        self._skipped = False

        # Reference count for dependent tests; safe to cleanup the test only
        # if it is zero
        self.ref_count = case.num_dependents

        # Test case has finished, but has not been waited for yet
        self.zombie = False

        # Timestamps for the start and finish phases of the pipeline
        self._timestamps = {}

        self._aborted = False

        # Performance logging
        self._perflogger = logging.null_logger
        self._perflog_compat = runtime.runtime().get_option(
            'logging/0/perflog_compat'
        )

    def duration(self, phase):
        # Treat pseudo-phases first
        if phase == 'compile_complete':
            t_start = 'compile_start'
            t_finish = 'compile_wait_finish'
        elif phase == 'run_complete':
            t_start = 'run_start'
            t_finish = 'wait_finish'
        elif phase == 'total':
            t_start = 'setup_start'
            t_finish = 'pipeline_end'
        else:
            t_start  = f'{phase}_start'
            t_finish = f'{phase}_finish'

        start = self._timestamps.get(t_start)
        if not start:
            return None

        finish = self._timestamps.get(t_finish)
        if not finish:
            finish = self._timestamps.get('pipeline_end')

        return finish - start

    def pipeline_timings(self, phases):
        def _tf(t):
            return f'{t:.3f}s' if t else 'n/a'

        msg = ''
        for phase in phases:
            if phase == 'compile_complete':
                msg += f"compile: {_tf(self.duration('compile_complete'))} "
            elif phase == 'run_complete':
                msg += f"run: {_tf(self.duration('run_complete'))} "
            else:
                msg += f"{phase}: {_tf(self.duration(phase))} "

        if msg:
            msg = msg[:-1]

        return msg

    def pipeline_timings_all(self):
        return self.pipeline_timings([
            'setup', 'compile_complete', 'run_complete',
            'sanity', 'performance', 'total'
        ])

    def pipeline_timings_basic(self):
        return self.pipeline_timings([
            'compile_complete', 'run_complete', 'total'
        ])

    @property
    def testcase(self):
        return self._case

    @property
    def check(self):
        return self._case.check

    @property
    def exc_info(self):
        return self._exc_info

    @property
    def failed(self):
        return (self._failed_stage is not None and
                not self._aborted and not self._skipped)

    @property
    def state(self):
        if self.failed:
            return 'fail'

        if self.skipped:
            return 'skip'

        states = {
            'startup': 'startup',
            'setup': 'ready_compile',
            'compile': 'compiling',
            'compile_wait': 'ready_run',
            'run': 'running',
            'run_wait': 'completing',
            'finalize': 'retired',
            'cleanup': 'completed',
        }
        return states[self._current_stage]

    @property
    def failed_stage(self):
        return self._failed_stage

    @property
    def succeeded(self):
        return self._current_stage in {'finalize', 'cleanup'}

    @property
    def completed(self):
        return self.failed or self.succeeded

    @property
    def aborted(self):
        return self._aborted

    @property
    def skipped(self):
        return self._skipped

    def _notify_listeners(self, callback_name):
        for l in self._listeners:
            callback = getattr(l, callback_name)
            callback(self)

    def _safe_call(self, fn, *args, **kwargs):
        class update_timestamps:
            '''Context manager to set the start and finish timestamps.'''

            # We use `this` to refer to the update_timestamps object, because
            # we don't want to masquerade the self argument of our containing
            # function
            def __enter__(this):
                if fn.__name__ not in ('poll',
                                       'run_complete',
                                       'compile_complete'):
                    stage = self._current_stage
                    self._timestamps[f'{stage}_start'] = time.time()

            def __exit__(this, exc_type, exc_value, traceback):
                stage = self._current_stage
                self._timestamps[f'{stage}_finish'] = time.time()
                self._timestamps['pipeline_end'] = time.time()

        if fn.__name__ not in ('poll', 'run_complete', 'compile_complete'):
            self._current_stage = fn.__name__

        try:
            with logging.logging_context(self.check) as logger:
                logger.debug(f'Entering stage: {self._current_stage}')
                with update_timestamps():
                    # Pick the configuration of the current partition
                    with runtime.temp_config(self.testcase.partition.fullname):
                        return fn(*args, **kwargs)
        except SkipTestError as e:
            if not self.succeeded:
                # Only skip a test if it hasn't finished yet;
                # This practically ignores skipping during the cleanup phase
                self.skip()
                raise TaskExit from e
        except ABORT_REASONS:
            self.fail()
            raise
        except BaseException as e:
            self.fail()
            raise TaskExit from e

    @logging.time_function
    def setup(self, *args, **kwargs):
        self.testcase.prepare()
        self._safe_call(self.check.setup, *args, **kwargs)
        self._notify_listeners('on_task_setup')

    @logging.time_function
    def compile(self):
        self._safe_call(self.check.compile)
        self._notify_listeners('on_task_compile')

    @logging.time_function
    def compile_wait(self):
        self._safe_call(self.check.compile_wait)

    @logging.time_function
    def run(self):
        self._safe_call(self.check.run)
        self._notify_listeners('on_task_run')

    @logging.time_function
    def run_complete(self):
        done = self._safe_call(self.check.run_complete)
        if done:
            self.zombie = True
            self._notify_listeners('on_task_exit')

        return done

    @logging.time_function
    def compile_complete(self):
        done = self._safe_call(self.check.compile_complete)
        if done:
            self._notify_listeners('on_task_compile_exit')

        return done

    @logging.time_function
    def run_wait(self):
        self._safe_call(self.check.run_wait)
        self.zombie = False

    @logging.time_function
    def sanity(self):
        self._safe_call(self.check.sanity)

    @logging.time_function
    def performance(self):
        self._perflogger = logging.getperflogger(self.check)
        self._safe_call(self.check.performance)

    @logging.time_function
    def finalize(self):
        try:
            jsonfile = os.path.join(self.check.stagedir, '.rfm_testcase.json')
            with open(jsonfile, 'w') as fp:
                jsonext.dump(self.check, fp, indent=2)
        except OSError as e:
            logging.getlogger().warning(
                f'could not dump test case {self.testcase}: {e}'
            )

        self._current_stage = 'finalize'
        self._notify_listeners('on_task_success')
        self._perflogger.log_performance(logging.INFO, self,
                                         multiline=self._perflog_compat)

    @logging.time_function
    def cleanup(self, *args, **kwargs):
        self._safe_call(self.check.cleanup, *args, **kwargs)

    def fail(self, exc_info=None):
        self._failed_stage = self._current_stage
        self._exc_info = exc_info or sys.exc_info()
        self._notify_listeners('on_task_failure')
        self._perflogger.log_performance(logging.INFO, self,
                                         multiline=self._perflog_compat)

    def skip(self, exc_info=None):
        self._skipped = True
        self._failed_stage = self._current_stage
        self._exc_info = exc_info or sys.exc_info()
        self._notify_listeners('on_task_skip')

    def abort(self, cause=None):
        if self.failed or self._aborted:
            return

        logging.getlogger().debug2('Aborting test case: {self.testcase!r}')
        exc = AbortTaskError()
        exc.__cause__ = cause
        self._aborted = True
        try:
            # FIXME: we should perhaps extend the RegressionTest interface
            # for supporting job cancelling
            if not self.zombie and self.check.job:
                self.check.job.cancel()
        except JobNotStartedError:
            self.fail((type(exc), exc, None))
        except BaseException:
            self.fail()
        else:
            self.fail((type(exc), exc, None))

    def info(self):
        '''Return an info string about this task.'''
        name = self.check.display_name
        hashcode = self.check.hashcode
        part = self.testcase.partition.fullname
        env  = self.testcase.environ.name
        return f'{name} /{hashcode} @{part}+{env}'


class TaskEventListener(abc.ABC):
    @abc.abstractmethod
    def on_task_setup(self, task):
        '''Called whenever the setup() method of a RegressionTask is called.'''

    @abc.abstractmethod
    def on_task_run(self, task):
        '''Called whenever the run() method of a RegressionTask is called.'''

    @abc.abstractmethod
    def on_task_compile(self, task):
        '''Called whenever the compile() method of a RegressionTask is
        called.'''

    @abc.abstractmethod
    def on_task_exit(self, task):
        '''Called whenever a RegressionTask finishes.'''

    @abc.abstractmethod
    def on_task_compile_exit(self, task):
        '''Called whenever a RegressionTask compilation phase finishes.'''

    @abc.abstractmethod
    def on_task_skip(self, task):
        '''Called whenever a RegressionTask is skipped.'''

    @abc.abstractmethod
    def on_task_failure(self, task):
        '''Called when a regression test has failed.'''

    @abc.abstractmethod
    def on_task_success(self, task):
        '''Called when a regression test has succeeded.'''


def _handle_sigterm(signum, frame):
    raise ForceExitError('received TERM signal')


class Runner:
    '''Responsible for executing a set of regression tests based on an
    execution policy.'''

    def __init__(self, policy, printer=None, max_retries=0,
                 max_failures=sys.maxsize):
        self._policy = policy
        self._printer = printer or PrettyPrinter()
        self._max_retries = max_retries
        self._stats = TestStats()
        self._policy.stats = self._stats
        self._policy.printer = self._printer
        self._policy.max_failures = max_failures
        signal.signal(signal.SIGTERM, _handle_sigterm)

    @property
    def max_failures(self):
        return self._max_failures

    @property
    def max_retries(self):
        return self._max_retries

    @property
    def policy(self):
        return self._policy

    @property
    def stats(self):
        return self._stats

    @logging.time_function
    def runall(self, testcases, restored_cases=None):
        num_checks = len({tc.check.unique_name for tc in testcases})
        self._printer.separator('short double line',
                                'Running %d check(s)' % num_checks)
        self._printer.timestamp('Started on', 'short double line')
        self._printer.info('')
        try:
            self._runall(testcases)
            if self._max_retries:
                restored_cases = restored_cases or []
                self._retry_failed(testcases + restored_cases)
        finally:
            # Print the summary line
            num_failures = len(self._stats.failed())
            num_completed = len(self._stats.completed())
            num_skipped = len(self._stats.skipped())
            num_tasks = len(self._stats.tasks())
            if num_failures > 0 or num_completed + num_skipped < num_tasks:
                status = 'FAILED'
            else:
                status = 'PASSED'

            total_run = len(testcases)
            total_completed = len(self._stats.completed(0))
            total_skipped = len(self._stats.skipped(0))
            self._printer.status(
                status,
                f'Ran {total_completed}/{total_run}'
                f' test case(s) from {num_checks} check(s) '
                f'({num_failures} failure(s), {total_skipped} skipped)',
                just='center'
            )
            self._printer.timestamp('Finished on', 'short double line')

    def _retry_failed(self, cases):
        rt = runtime.runtime()
        failures = self._stats.failed()
        while (failures and rt.current_run < self._max_retries):
            num_failed_checks = len({tc.check.unique_name for tc in failures})
            rt.next_run()

            self._printer.separator(
                'short double line',
                'Retrying %d failed check(s) (retry %d/%d)' %
                (num_failed_checks, rt.current_run, self._max_retries)
            )

            # Clone failed cases and rebuild dependencies among them
            failed_cases = [t.testcase.clone() for t in failures]
            cases_graph, _ = dependencies.build_deps(failed_cases, cases)
            failed_cases = dependencies.toposort(cases_graph, is_subgraph=True)
            self._runall(failed_cases)
            failures = self._stats.failed()

    def _runall(self, testcases):
        def print_separator(check, prefix):
            self._printer.separator(
                'short single line',
                '%s %s (%s)' % (prefix, check.unique_name, check.descr)
            )

        self._printer.separator('short single line',
                                'start processing checks')
        self._policy.enter()
        self._printer.reset_progress(len(testcases))
        for t in testcases:
            self._policy.runcase(t)

        self._policy.exit()
        self._printer.separator('short single line',
                                'all spawned checks have finished\n')


class ExecutionPolicy(abc.ABC):
    '''Base abstract class for execution policies.

    An execution policy implements the regression check pipeline.'''

    def __init__(self):
        # Options controlling the check execution
        self.skip_sanity_check = False
        self.skip_performance_check = False
        self.keep_stage_files = False
        self.only_environs = None
        self.printer = None

        # Local scheduler for running forced local jobs
        self.local_scheduler = LocalJobScheduler()

        # Scheduler options
        self.sched_flex_alloc_nodes = None
        self.sched_options = []

        # Task event listeners
        self.task_listeners = []
        self.stats = None

    def enter(self):
        self._num_failed_tasks = 0

    def exit(self):
        pass

    @abc.abstractmethod
    def runcase(self, case):
        '''Run a test case.'''



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import contextlib
import copy
import fnmatch
import functools
import itertools
import json
import jsonschema
import os
import re
import socket

import reframe
import reframe.core.settings as settings
import reframe.utility as util
from reframe.core.environments import normalize_module_list
from reframe.core.exceptions import ConfigError, ReframeFatalError
from reframe.core.logging import getlogger
from reframe.utility import ScopedDict


def _match_option(opt, opt_map):
    if isinstance(opt, list):
        opt = '/'.join(opt)

    if opt in opt_map:
        return opt_map[opt]

    for k, v in opt_map.items():
        if fnmatch.fnmatchcase(opt, k):
            return v

    raise KeyError(opt)


def _normalize_syntax(conv):
    '''Normalize syntax for options accepting multiple syntaxes'''

    def _do_normalize(fn):

        @functools.wraps(fn)
        def _get(site_config, option, *args, **kwargs):
            ret = fn(site_config, option, *args, **kwargs)
            if option is None:
                return ret

            for opt_patt, norm_fn in conv.items():
                if re.match(opt_patt, option):
                    ret = norm_fn(ret)
                    break

            return ret

        return _get

    return _do_normalize


def _hostname(use_fqdn, use_xthostname):
    '''Return hostname'''
    if use_xthostname:
        try:
            xthostname_file = '/etc/xthostname'
            getlogger().debug(f'Trying {xthostname_file!r}...')
            with open(xthostname_file) as fp:
                return fp.read()
        except OSError as e:
            '''Log the error and continue to the next method'''
            getlogger().debug(f'Failed to read {xthostname_file!r}')

    if use_fqdn:
        getlogger().debug('Using FQDN...')
        return socket.getfqdn()

    getlogger().debug('Using standard hostname...')
    return socket.gethostname()


class _SiteConfig:
    def __init__(self):
        self._site_config = None
        self._sources = []
        self._subconfigs = {}
        self._local_system = None
        self._sticky_options = {}
        self._autodetect_meth = 'hostname'
        self._autodetect_opts = {
            'hostname': {
                'use_fqdn': False,
                'use_xthostname': False,
            }
        }
        self._definitions = {
            'systems': {},
            'partitions': {},
            'environments': {},
            'modes': {}
        }

        # Open and store the JSON schema for later validation
        schema_filename = os.path.join(reframe.INSTALL_PREFIX, 'reframe',
                                       'schemas', 'config.json')
        with open(schema_filename) as fp:
            try:
                self._schema = json.loads(fp.read())
            except json.JSONDecodeError as e:
                raise ReframeFatalError(
                    f'invalid configuration schema: {schema_filename!r}'
                ) from e

    def _update_system_defs(self, config, filename):
        for sys_entry in config:
            sys_name = sys_entry['name']
            if sys_name in self._definitions['systems']:
                fname = self._definitions['systems'][sys_name]
                getlogger().warning(f'redefinition of system {sys_name!r}: '
                                    f'already defined in {fname!r}')

            self._definitions['systems'][sys_name] = filename
            for part_entry in sys_entry['partitions']:
                part_name = sys_name + ':' + part_entry['name']
                if part_name in self._definitions['partitions']:
                    fname = self._definitions['partitions'][part_name]
                    getlogger().warning(
                        f'redefinition of partition {part_name!r}: '
                        f'already defined in {fname!r}'
                    )
                self._definitions['partitions'][part_name] = filename

    def _update_environment_defs(self, config, filename):
        for env_entry in config:
            target_systems = env_entry.get('target_systems', '*')
            for target in target_systems:
                env_name = target + ':' + env_entry['name']
                if env_name in self._definitions['environments']:
                    fname = self._definitions['environments'][env_name]
                    getlogger().warning(
                        f'redefinition of environment {env_name!r}: '
                        f'already defined in {fname!r}'
                    )

                self._definitions['environments'][env_name] = filename

    def _update_mode_defs(self, config, filename):
        for mode_entry in config:
            target_systems = mode_entry.get('target_systems', '*')
            for target in target_systems:
                mode_name = target + ':' + mode_entry['name']
                if mode_name in self._definitions['modes']:
                    fname = self._definitions['modes'][mode_name]
                    getlogger().warning(f'redefinition of mode {mode_name!r}: '
                                        f'already defined in {fname!r}')

                self._definitions['modes'][mode_name] = filename

    def _update_defs(self, site_config, filename):
        for secname, config in site_config.items():
            if secname == 'systems':
                self._update_system_defs(config, filename)
            elif secname == 'environments':
                self._update_environment_defs(config, filename)
            elif secname == 'modes':
                self._update_mode_defs(config, filename)

    def _merge_config_sections(self, target, other):
        '''Merge `other` section into `target`.

        :returns: the merged section
        '''
        # Index the sections by the target_systems
        options_by_system = {}
        for entry in target + other:
            for key in entry.pop('target_systems', ['*']):
                options_by_system.setdefault(key, [])
                options_by_system[key].append(entry)

        # Now merge the options
        ret = []
        for system, optionset in options_by_system.items():
            entry = functools.reduce(
                lambda l, r: l.update(r) or l, optionset
            )
            entry['target_systems'] = [system]
            ret.append(entry)

        return ret

    def update_config(self, config, filename):
        self._sources.append(filename)
        self._update_defs(config, filename)
        nc = copy.deepcopy(config)
        if self._site_config is None:
            self._site_config = nc
            return self

        mergeable_sections = ('general', 'logging', 'schedulers')
        for sec in nc.keys():
            if sec not in self._site_config:
                self._site_config[sec] = nc[sec]
            elif sec in mergeable_sections:
                self._site_config[sec] = self._merge_config_sections(
                    self._site_config[sec], nc[sec]
                )
            else:
                if sec == 'systems':
                    # Systems have to be inserted in the beginning of the list,
                    # since they are selected by the first matching entry in
                    # `hostnames`.
                    self._site_config[sec] = nc[sec] + self._site_config[sec]
                else:
                    self._site_config[sec] += nc[sec]

    def _pick_config(self):
        if self._local_system:
            return self._subconfigs[self._local_system]
        else:
            return self._site_config

    def __repr__(self):
        return (f'{type(self).__name__}(site_config={self._site_config!r}, '
                f'sources={self._sources!r})')

    def __str__(self):
        return json.dumps(self._pick_config(), indent=2)

    # Delegate everything to either the original config or to the reduced one
    # if a system is selected

    def __iter__(self):
        return iter(self._pick_config())

    def __getitem__(self, key):
        return self._pick_config()[key]

    def __getattr__(self, attr):
        return getattr(self._pick_config(), attr)

    def set_autodetect_meth(self, method, **opts):
        self._autodetect_meth = method
        try:
            self._autodetect_opts[method].update(opts)
        except KeyError:
            raise ConfigError(
                f'unknown auto-detection method: {method!r}'
            ) from None

    @property
    def schema(self):
        '''Configuration schema'''
        return self._schema

    def add_sticky_option(self, option, value):
        self._sticky_options[option] = value

    def remove_sticky_option(self, option):
        self._sticky_options.pop(option, None)

    def is_sticky_option(self, option):
        return option in self._sticky_options

    @_normalize_syntax({'.*/.*modules$': normalize_module_list})
    def get(self, option, default=None):
        '''Retrieve value of option.

        If the option cannot be retrieved, ``default`` will be returned.
        '''

        # Options may not start with a slash
        if not option or option[0] == '/':
            return default

        # Remove trailing /
        if option[-1] == '/':
            option = option[:-1]

        # Convert any indices to integers
        prepared_option = []
        for opt in option.split('/'):
            try:
                opt = int(opt)
            except ValueError:
                pass

            prepared_option.append(opt)

        # Walk through the option path constructing a default key at the same
        # time for looking it up in the defaults or the sticky options
        default_key = []
        value = self._pick_config()
        option_path_invalid = False
        for x in prepared_option:
            if option_path_invalid:
                # Just go through the rest of elements and construct the key
                # trivially
                if not isinstance(x, int) and x[0] != '@':
                    default_key.append(x)

                continue

            if isinstance(x, int) or x[0] == '@':
                # We are in an addressable element; move forward in the path,
                # without adding the component to the default_key
                if isinstance(x, int):
                    # Element addressable by index number
                    try:
                        value = value[x]
                    except IndexError:
                        option_path_invalid = True
                else:
                    # Element addressable by name
                    x, found = x[1:], False
                    for obj in value:
                        if obj['name'] == x:
                            value, found = obj, True
                            break

                    if not found:
                        option_path_invalid = True

                continue

            if 'type' in value:
                default_key.append(value['type'] + '_' + x)
            else:
                default_key.append(x)

            try:
                value = value[x]
            except (IndexError, KeyError, TypeError):
                option_path_invalid = True

        default_key = '/'.join(default_key)
        try:
            # If a sticky option exists, return that value
            return _match_option(default_key, self._sticky_options)
        except KeyError:
            pass

        if option_path_invalid:
            # Try the default and return
            try:
                return _match_option(default_key, self._schema['defaults'])
            except KeyError:
                return default

        return value

    @property
    def sources(self):
        return self._sources

    @property
    def subconfig_system(self):
        return self._local_system

    def load_config_python(self, filename):
        try:
            mod = util.import_module_from_file(filename)
        except ImportError as e:
            # import_module_from_file() may raise an ImportError if the
            # configuration file is under ReFrame's top-level directory
            raise ConfigError(
                f"could not load Python configuration file: '{filename}'"
            ) from e

        if hasattr(mod, 'settings'):
            # Looks like an old style config
            raise ConfigError(
                f"the syntax of the configuration file {filename!r} "
                f"is no longer supported"
            )

        mod = util.import_module_from_file(filename)
        if not hasattr(mod, 'site_configuration'):
            raise ConfigError(
                f"not a valid Python configuration file: '{filename}'"
            )

        self.update_config(mod.site_configuration, filename)

    def load_config_json(self, filename):
        with open(filename) as fp:
            try:
                config = json.loads(fp.read())
            except json.JSONDecodeError as e:
                raise ConfigError(
                    f"invalid JSON syntax in configuration file '{filename}'"
                ) from e

        self.update_config(config, filename)

    def _detect_system(self):
        getlogger().debug(
            f'Detecting system using method: {self._autodetect_meth!r}'
        )
        hostname = _hostname(
            self._autodetect_opts[self._autodetect_meth]['use_fqdn'],
            self._autodetect_opts[self._autodetect_meth]['use_xthostname'],
        )
        getlogger().debug(f'Retrieved hostname: {hostname!r}')
        getlogger().debug(f'Looking for a matching configuration entry')
        for system in self._site_config['systems']:
            for patt in system['hostnames']:
                if re.match(patt, hostname):
                    sysname = system['name']
                    getlogger().debug(
                        f'Configuration found: picking system {sysname!r}'
                    )
                    return sysname

        raise ConfigError(f"could not find a configuration entry "
                          f"for the current system: '{hostname}'")

    def validate(self):
        site_config = self._pick_config()
        try:
            jsonschema.validate(site_config, self._schema)
        except jsonschema.ValidationError as e:
            raise ConfigError(f"could not validate configuration files: "
                              f"'{self._sources}'") from e

        def _warn_variables(config, opt_path):
            opt_path = '/'.join(opt_path + ['variables'])
            if 'env_vars' in config and 'variables' in config:
                getlogger().warning(
                    f"configuration option {opt_path!r}: "
                    f"both 'env_vars' and 'variables' are defined; "
                    f"'variables' will be ignored"
                )
            elif 'variables' in config:
                getlogger().warning(
                    f"configuration option {opt_path!r}: "
                    f"'variables' is deprecated; please use 'env_vars' instead"
                )
                config['env_vars'] = config['variables']

        # Warn about the deprecated `variables` and convert them internally to
        # `env_vars`
        for system in self._site_config['systems']:
            sysname = system['name']
            opt_path = ['systems', f'@{sysname}']
            _warn_variables(system, opt_path)
            for part in system['partitions']:
                partname = part['name']
                opt_path += ['partitions', f'@{partname}']
                _warn_variables(part, opt_path)
                for i, cp in enumerate(part.get('container_platforms', [])):
                    opt_path += ['container_platforms', str(i)]
                    _warn_variables(cp, opt_path)
                    opt_path.pop()
                    opt_path.pop()

                opt_path.pop()
                opt_path.pop()

        for env in self._site_config['environments']:
            envname = env['name']
            opt_path = ['environments', f'@{envname}']
            _warn_variables(env, opt_path)

    def select_subconfig(self, system_fullname=None,
                         ignore_resolve_errors=False):
        # First look for the current subconfig in the cache; if not found,
        # generate it and cache it
        system_fullname = system_fullname or self._detect_system()
        getlogger().debug2(f'Selecting subconfig for {system_fullname!r}')

        self._local_system = system_fullname
        if system_fullname in self._subconfigs:
            return

        try:
            system_name, part_name = system_fullname.split(':', maxsplit=1)
        except ValueError:
            # system_name does not have a partition
            system_name, part_name = system_fullname, None

        # Start from a fresh copy of the site_config, because we will be
        # modifying it
        site_config = copy.deepcopy(self._site_config)
        local_config = {}
        systems = list(
            filter(lambda x: x['name'] == system_name, site_config['systems'])
        )
        if not systems:
            raise ConfigError(
                f"could not find a configuration entry "
                f"for the requested system: '{system_name}'"
            )

        if part_name is not None:
            # Filter out also partitions
            systems[0]['partitions'] = list(
                filter(lambda x: x['name'] == part_name,
                       systems[0]['partitions'])
            )

        if not systems[0]['partitions']:
            raise ConfigError(
                f"could not find a configuration entry "
                f"for the requested system/partition combination: "
                f"'{system_name}:{part_name}'"
            )

        # Create local configuration for the current or the requested system
        local_config['systems'] = systems
        for name, section in site_config.items():
            if name == 'systems':
                # The systems sections has already been treated
                continue

            # Convert section to a scoped dict that will handle correctly and
            # transparently the system/partition resolution
            scoped_section = ScopedDict()
            unnamed_objects = False
            for obj in section:
                target_systems = obj.get(
                    'target_systems',
                    _match_option(f'{name}/target_systems',
                                  self._schema['defaults'])
                )
                try:
                    key = obj['name']
                    for t in target_systems:
                        scoped_section[f'{t}:{key}'] = obj
                except KeyError:
                    unnamed_objects = True
                    for k, v in obj.items():
                        if k != 'target_systems':
                            for t in target_systems:
                                scoped_section[f'{t}:{name}/{k}'] = v

            if unnamed_objects:
                # We need to merge all the objects of the section into a
                # single one based on the selected system
                uniq_obj = {}
                for obj in section:
                    for k, v in obj.items():
                        with contextlib.suppress(KeyError):
                            uniq_obj[k] = scoped_section[
                                f'{system_fullname}:{name}/{k}'
                            ]

                local_config[name] = [uniq_obj]
            else:
                unique_keys = set()
                for obj in section:
                    key = obj['name']
                    if key in unique_keys:
                        continue

                    unique_keys.add(key)
                    try:
                        val = scoped_section[f'{system_fullname}:{key}']
                    except KeyError:
                        pass
                    else:
                        local_config.setdefault(name, [])
                        local_config[name].append(val)

        required_sections = self._schema['required']
        for name in required_sections:
            if name not in local_config.keys():
                if not ignore_resolve_errors:
                    raise ConfigError(
                        f"section '{name}' not defined "
                        f"for system '{system_fullname}'"
                    )

        # Check that handlers$ are defined for the current system
        if 'handlers$' not in local_config['logging'][0]:
            raise ConfigError(f"'logging/handlers$' are not defined "
                              f"for system {system_fullname!r}")

        # Check that all environments defined by the system are defined for
        # the current system
        if not ignore_resolve_errors:
            sys_environs = {
                *itertools.chain(*(p['environs']
                                   for p in systems[0]['partitions']))
            }
            found_environs = {
                e['name'] for e in local_config['environments']
            }
            undefined_environs = sys_environs - found_environs
            if undefined_environs:
                env_descr = ', '.join(f"'{e}'" for e in undefined_environs)
                raise ConfigError(
                    f"environments {env_descr} "
                    f"are not defined for '{system_fullname}'"
                )

        self._subconfigs[system_fullname] = local_config


def find_config_files(config_path=None, config_file=None):
    res = []
    if config_path:
        for p in config_path:
            if os.path.exists(p + '/settings.py'):
                res.append(p + '/settings.py')
            elif os.path.exists(p + '/settings.json'):
                res.append(p + '/settings.json')
            else:
                getlogger().debug(f"No 'settings.py' or 'settings.json' "
                                  f"found in {p!r}, path will be ignored")

    if config_file:
        for f in config_file:
            # If the user sets RFM_CONFIG_FILES=:conf1:conf2 the list will
            # include one empty string in the beginning
            if f == '':
                res = []
            elif f.startswith(':'):
                res = [f[1:]]
            else:
                res.append(f)

    return res


def load_config(*filenames):
    ret = _SiteConfig()
    getlogger().debug('Loading the generic configuration')
    ret.update_config(settings.site_configuration, '<builtin>')
    for f in filenames:
        getlogger().debug(f'Loading configuration file: {filenames!r}')
        _, ext = os.path.splitext(f)
        if ext == '.py':
            ret.load_config_python(f)
        elif ext == '.json':
            ret.load_config_json(f)
        else:
            raise ConfigError(f"unknown configuration file type: '{f}'")

    return ret



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import collections
import os

import reframe.core.fields as fields
import reframe.utility as util
import reframe.utility.jsonext as jsonext
import reframe.utility.typecheck as typ
from reframe.core.warnings import user_deprecation_warning


def normalize_module_list(modules):
    '''Normalize module list.

    :meta private:
    '''
    ret = []
    for m in modules:
        if isinstance(m, str):
            ret.append({'name': m, 'collection': False, 'path': None})
        else:
            ret.append(m)

    return ret


class Environment(jsonext.JSONSerializable):
    '''This class abstracts away an environment to run regression tests.

    It is simply a collection of modules to be loaded and environment variables
    to be set when this environment is loaded by the framework.

    .. warning::
       Users may not create :class:`Environment` objects directly.
    '''

    def __init__(self, name, modules=None, env_vars=None,
                 extras=None, features=None):
        modules = modules or []
        env_vars = env_vars or []
        self._name = name
        self._modules = normalize_module_list(modules)
        self._module_names = [m['name'] for m in self._modules]

        # Convert values of env_vars to strings before storing
        if isinstance(env_vars, dict):
            env_vars = env_vars.items()

        self._env_vars = collections.OrderedDict()
        for k, v in env_vars:
            self._env_vars[k] = str(v)

        self._extras = extras or {}
        self._features = features or []

    @property
    def name(self):
        '''The name of this environment.

        :type: :class:`str`
        '''
        return self._name

    @property
    def modules(self):
        '''The modules associated with this environment.

        :type: :class:`List[str]`
        '''
        return util.SequenceView(self._module_names)

    @property
    def modules_detailed(self):
        '''A view of the modules associated with this environment in a detailed
        format.

        Each module is represented as a dictionary with the following
        attributes:

        - ``name``: the name of the module.
        - ``collection``: :class:`True` if the module name refers to a module
          collection.

        :type: :class:`List[Dict[str, object]]`

        .. versionadded:: 3.3

        '''

        return util.SequenceView(self._modules)

    @property
    def env_vars(self):
        '''The environment variables associated with this environment.

        :type: :class:`OrderedDict[str, str]`

        .. versionadded:: 4.0.0
        '''
        return util.MappingView(self._env_vars)

    @property
    def variables(self):
        '''The environment variables associated with this environment.

        .. deprecated:: 4.0.0
           Please :attr:`env_vars` instead.
        '''
        user_deprecation_warning("the 'variables' attribute is deprecated; "
                                 "please use the 'env_vars' instead")
        return util.MappingView(self._env_vars)

    @property
    def extras(self):
        '''User defined properties specified in the configuration.

        .. versionadded:: 3.9.1

        :type: :class:`Dict[str, object]`
        '''

        return self._extras

    @property
    def features(self):
        '''Used defined features specified in the configuration.

        .. versionadded:: 3.11.0

        :type: :class:`List[str]`
        '''
        return self._features

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return NotImplemented

        if (self.name != other.name or
            set(self.modules) != set(other.modules)):
            return False

        # Env. variables are checked against their string representation
        for kv0, kv1 in zip(self.env_vars.items(),
                            other.env_vars.items()):
            k0, v0 = kv0
            k1, v1 = kv1
            if k0 != k1 or str(v0) != str(v1):
                return False

        return True

    def __str__(self):
        return self.name

    def __repr__(self):
        return (f'{type(self).__name__}('
                f'name={self._name!r}, '
                f'modules={self._modules!r}, '
                f'env_vars={list(self._env_vars.items())!r}, '
                f'extras={self._extras!r}, features={self._features!r})')


class _EnvironmentSnapshot(Environment):
    '''An environment snapshot.'''

    def __init__(self, name='env_snapshot'):
        super().__init__(name, [], os.environ.items())

    def restore(self):
        '''Restore this environment snapshot.'''
        os.environ.clear()
        os.environ.update(self._env_vars)

    def __eq__(self, other):
        if not isinstance(other, Environment):
            return NotImplemented

        # Order of env. variables is not important when comparing snapshots
        for k, v in self.env_vars.items():
            if other.env_vars[k] != v:
                return False

        return self.name == other.name


def snapshot():
    '''Create an environment snapshot

    :returns: An instance of :class:`_EnvironmentSnapshot`.
    '''
    return _EnvironmentSnapshot()


class ProgEnvironment(Environment):
    '''A class representing a programming environment.

    This type of environment adds also properties for retrieving the compiler
    and compilation flags.

    .. warning::
       Users may not create :class:`ProgEnvironment` objects directly.
    '''

    _cc = fields.TypedField(str)
    _cxx = fields.TypedField(str)
    _ftn = fields.TypedField(str)
    _cppflags = fields.TypedField(typ.List[str])
    _cflags = fields.TypedField(typ.List[str])
    _cxxflags = fields.TypedField(typ.List[str])
    _fflags = fields.TypedField(typ.List[str])
    _ldflags = fields.TypedField(typ.List[str])

    def __init__(self,
                 name,
                 modules=None,
                 env_vars=None,
                 extras=None,
                 features=None,
                 cc='cc',
                 cxx='CC',
                 ftn='ftn',
                 nvcc='nvcc',
                 cppflags=None,
                 cflags=None,
                 cxxflags=None,
                 fflags=None,
                 ldflags=None,
                 **kwargs):
        super().__init__(name, modules, env_vars, extras, features)
        self._cc = cc
        self._cxx = cxx
        self._ftn = ftn
        self._nvcc = nvcc
        self._cppflags = cppflags or []
        self._cflags   = cflags   or []
        self._cxxflags = cxxflags or []
        self._fflags   = fflags   or []
        self._ldflags  = ldflags  or []

    @property
    def cc(self):
        '''The C compiler of this programming environment.

        :type: :class:`str`
        '''
        return self._cc

    @property
    def cxx(self):
        '''The C++ compiler of this programming environment.

        :type: :class:`str`
        '''
        return self._cxx

    @property
    def ftn(self):
        '''The Fortran compiler of this programming environment.

        :type: :class:`str`
        '''
        return self._ftn

    @property
    def cppflags(self):
        '''The preprocessor flags of this programming environment.

        :type: :class:`List[str]`
        '''
        return self._cppflags

    @property
    def cflags(self):
        '''The C compiler flags of this programming environment.

        :type: :class:`List[str]`
        '''
        return self._cflags

    @property
    def cxxflags(self):
        '''The C++ compiler flags of this programming environment.

        :type: :class:`List[str]`
        '''
        return self._cxxflags

    @property
    def fflags(self):
        '''The Fortran compiler flags of this programming environment.

        :type: :class:`List[str]`
        '''
        return self._fflags

    @property
    def ldflags(self):
        '''The linker flags of this programming environment.

        :type: :class:`List[str]`
        '''
        return self._ldflags

    @property
    def nvcc(self):
        return self._nvcc



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import abc
import re
import semver


parse = semver.VersionInfo.parse


class _ValidatorImpl(abc.ABC):
    '''Abstract base class for the validation of version ranges.'''
    @abc.abstractmethod
    def validate(version):
        pass


class _IntervalValidator(_ValidatorImpl):
    '''Class for the validation of version intervals.

    This class takes an interval of versions "v1..v2" and its method
    ``validate`` returns ``True`` if a given version string is inside
    the interval including the endpoints.
    '''

    def __init__(self, condition):
        try:
            min_version_str, max_version_str = condition.split('..')
        except ValueError:
            raise ValueError("invalid format of version interval: %s" %
                             condition) from None

        if min_version_str and max_version_str:
            self._min_version = parse(min_version_str)
            self._max_version = parse(max_version_str)
        else:
            raise ValueError("missing bound on version interval %s" %
                             condition)

    def validate(self, version):
        version = parse(version)
        return ((version >= self._min_version) and
                (version <= self._max_version))


class _RelationalValidator(_ValidatorImpl):
    '''Class for the validation of Boolean relations of versions.

    This class takes a Boolean relation of versions with the form
    ``<bool_operator><version>``, and its method ``validate`` returns
    ``True`` if a given version string satisfies the relation.
    '''

    def __init__(self, condition):
        self._op_actions = {
            ">":  lambda x, y: x > y,
            ">=": lambda x, y: x >= y,
            "<":  lambda x, y: x < y,
            "<=": lambda x, y: x <= y,
            "==": lambda x, y: x == y,
            "!=": lambda x, y: x != y,
        }
        cond_match = re.match(r'(\W{0,2})(\S+)', condition)
        if not cond_match:
            raise ValueError("invalid condition: '%s'" % condition)

        self._ref_version = parse(cond_match.group(2))
        op = cond_match.group(1)
        if not op:
            op = '=='

        if op not in self._op_actions.keys():
            raise ValueError("invalid boolean operator: '%s'" % op)
        else:
            self._operator = op

    def validate(self, version):
        do_validate = self._op_actions[self._operator]
        return do_validate(parse(version), self._ref_version)


class VersionValidator:
    '''Class factory for the validation of version ranges.'''
    def __new__(cls, condition):
        if '..' in condition:
            return _IntervalValidator(condition)
        else:
            return _RelationalValidator(condition)



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import semver
import re

import reframe.utility.osext as osext
from reframe.core.backends import register_launcher
from reframe.core.launchers import JobLauncher
from reframe.core.logging import getlogger
from reframe.utility import seconds_to_hms


@register_launcher('srun')
class SrunLauncher(JobLauncher):
    def __init__(self):
        self.options = []
        self.use_cpus_per_task = True
        try:
            out = osext.run_command('srun --version')
            match = re.search('slurm (\d+)\.(\d+)\.(\d+)', out.stdout)
            if match:
                # We cannot pass to semver strings like 22.05.1 directly
                # because it is not a valid version string for semver. We
                # need to remove all the leading zeros.
                slurm_version = (
                    semver.VersionInfo(
                        match.group(1), match.group(2), match.group(3)
                    )
                )
                if slurm_version < semver.VersionInfo(22, 5, 0):
                    self.use_cpus_per_task = False
            else:
                getlogger().warning(
                    'could not get version of Slurm, --cpus-per-task will be '
                    'set according to the num_cpus_per_task attribute'
                )
        except Exception:
            getlogger().warning(
                'could not get version of Slurm, --cpus-per-task will be set '
                'according to the num_cpus_per_task attribute'
            )

    def command(self, job):
        ret = ['srun']
        if self.use_cpus_per_task and job.num_cpus_per_task:
            ret.append(f'--cpus-per-task={job.num_cpus_per_task}')

        return ret


@register_launcher('ibrun')
class IbrunLauncher(JobLauncher):
    '''TACC's custom parallel job launcher.'''

    def command(self, job):
        return ['ibrun']


@register_launcher('upcrun')
class UpcrunLauncher(JobLauncher):
    '''Launcher for UPC applications.'''

    def command(self, job):
        cmd = ['upcrun']
        if job.num_tasks_per_node:
            num_nodes = job.num_tasks // job.num_tasks_per_node
            cmd += ['-N', str(num_nodes)]

        cmd += ['-n', str(job.num_tasks)]
        return cmd


@register_launcher('upcxx-run')
class UpcxxrunLauncher(JobLauncher):
    '''Launcher for UPC++ applications.'''

    def command(self, job):
        cmd = ['upcxx-run']
        if job.num_tasks_per_node:
            num_nodes = job.num_tasks // job.num_tasks_per_node
            cmd += ['-N', str(num_nodes)]

        cmd += ['-n', str(job.num_tasks)]
        return cmd


@register_launcher('alps')
class AlpsLauncher(JobLauncher):
    def command(self, job):
        cmd = ['aprun', '-n', str(job.num_tasks)]
        if job.num_tasks_per_node:
            cmd += ['-N', str(job.num_tasks_per_node)]

        if job.num_cpus_per_task:
            cmd += ['-d', str(job.num_cpus_per_task)]

        if job.use_smt:
            cmd += ['-j', '0']

        return cmd


@register_launcher('mpirun')
class MpirunLauncher(JobLauncher):
    def command(self, job):
        return ['mpirun', '-np', str(job.num_tasks)]


@register_launcher('mpiexec')
class MpiexecLauncher(JobLauncher):
    def command(self, job):
        return ['mpiexec', '-n', str(job.num_tasks)]


@register_launcher('srunalloc')
class SrunAllocationLauncher(JobLauncher):
    def command(self, job):
        ret = ['srun']
        if job.name:
            ret += ['--job-name=%s' % job.name]

        if job.time_limit:
            h, m, s = seconds_to_hms(job.time_limit)
            ret += ['--time=%d:%d:%d' % (h, m, s)]

        if job.stdout:
            ret += ['--output=%s' % job.stdout]

        if job.stderr:
            ret += ['--error=%s' % job.stderr]

        if job.num_tasks:
            ret += ['--ntasks=%s' % str(job.num_tasks)]

        if job.num_tasks_per_node:
            ret += ['--ntasks-per-node=%s' % str(job.num_tasks_per_node)]

        if job.num_tasks_per_core:
            ret += ['--ntasks-per-core=%s' % str(job.num_tasks_per_core)]

        if job.num_tasks_per_socket:
            ret += ['--ntasks-per-socket=%s' % str(job.num_tasks_per_socket)]

        if job.num_cpus_per_task:
            ret += ['--cpus-per-task=%s' % str(job.num_cpus_per_task)]

        if job.exclusive_access:
            ret += ['--exclusive']

        if job.use_smt is not None:
            hint = 'multithread' if job.use_smt else 'nomultithread'
            ret += ['--hint=%s' % hint]

        for opt in job.options + job.cli_options:
            if opt.startswith('#'):
                continue

            ret.append(opt)

        return ret


@register_launcher('lrun')
class LrunLauncher(JobLauncher):
    '''LLNL's custom parallel job launcher'''

    def command(self, job):
        num_tasks_per_node = job.num_tasks_per_node or 1
        num_nodes = job.num_tasks // num_tasks_per_node
        return ['lrun', '-N', str(num_nodes),
                '-T', str(num_tasks_per_node)]


@register_launcher('lrun-gpu')
class LrungpuLauncher(LrunLauncher):
    '''LLNL's custom parallel job launcher w/ CUDA aware Spectum MPI'''

    def command(self, job):
        return super().command(job) + ['-M "-gpu"']



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import abc

import reframe.core.fields as fields
import reframe.utility as util
import reframe.utility.typecheck as typ


_STAGEDIR_MOUNT = '/rfm_workdir'


class ContainerPlatform(abc.ABC):
    '''The abstract base class of any container platform.'''

    #: The default mount location of the test case stage directory inside the
    #: container

    #: The container image to be used for running the test.
    #:
    #: :type: :class:`str` or :class:`None`
    #: :default: :class:`None`
    image = fields.TypedField(str, type(None))

    #: The command to be executed within the container.
    #:
    #: If no command is given, then the default command of the corresponding
    #: container image is going to be executed.
    #:
    #: .. versionadded:: 3.5.0
    #:    Changed the attribute name from `commands` to `command` and its type
    #:    to a string.
    #:
    #: :type: :class:`str` or :class:`None`
    #: :default: :class:`None`
    command = fields.TypedField(str, type(None))

    #: Pull the container image before running.
    #:
    #: This does not have any effect for the `Singularity` container platform.
    #:
    #: .. versionadded:: 3.5
    #:
    #: :type: :class:`bool`
    #: :default: ``True``
    pull_image = fields.TypedField(bool)

    #: List of mount point pairs for directories to mount inside the container.
    #:
    #: Each mount point is specified as a tuple of
    #: ``(/path/in/host, /path/in/container)``. The stage directory of the
    #: ReFrame test is always mounted under ``/rfm_workdir`` inside the
    #: container, independelty of this field.
    #:
    #: :type: :class:`list[tuple[str, str]]`
    #: :default: ``[]``
    mount_points = fields.TypedField(typ.List[typ.Tuple[str, str]])

    #: Additional options to be passed to the container runtime when executed.
    #:
    #: :type: :class:`list[str]`
    #: :default: ``[]``
    options = fields.TypedField(typ.List[str])

    #: The working directory of ReFrame inside the container.
    #:
    #: This is the directory where the test's stage directory is mounted inside
    #: the container. This directory is always mounted regardless if
    #: :attr:`mount_points` is set or not.
    #:
    #: :type: :class:`str`
    #: :default: ``/rfm_workdir``
    #:
    #: .. versionchanged:: 3.12.0
    #:    This attribute is no more deprecated.
    workdir = fields.TypedField(str, type(None))

    def __init__(self):
        self.image = None
        self.command = None
        self.workdir = _STAGEDIR_MOUNT
        self.mount_points  = []
        self.options = []
        self.pull_image = True

    @abc.abstractmethod
    def emit_prepare_commands(self, stagedir):
        '''Returns commands for preparing this container for running.

        Such a command could be for pulling the container image from a
        repository.

        .. note:

            This method is relevant only to developers of new container
            platform backends.

        :meta private:

        :arg stagedir: The stage directory of the test.
        '''

    @abc.abstractmethod
    def launch_command(self, stagedir):
        '''Returns the command for running :attr:`commands` with this container
        platform.

        .. note:
            This method is relevant only to developers of new container
            platforms.

        :meta private:

        :arg stagedir: The stage directory of the test.
        '''

    @classmethod
    def create(cls, name):
        '''Factory method to create a new container by name.'''
        name = name.capitalize()
        try:
            return globals()[name]()
        except KeyError:
            raise ValueError(f'unknown container platform: {name}') from None

    @classmethod
    def create_from(cls, name, other):
        new = cls.create(name)
        new.image = other.image
        new.command = other.command
        new.mount_points = other.mount_points
        new.options = other.options
        new.pull_image = other.pull_image
        new.workdir = other.workdir
        return new

    @property
    def name(self):
        return type(self).__name__

    def __str__(self):
        return self.name

    def __rfm_json_encode__(self):
        return str(self)


class Docker(ContainerPlatform):
    '''Container platform backend for running containers with `Docker
    <https://www.docker.com/>`__.'''

    def emit_prepare_commands(self, stagedir):
        return [f'docker pull {self.image}'] if self.pull_image else []

    def launch_command(self, stagedir):
        super().launch_command(stagedir)
        mount_points = self.mount_points + [(stagedir, _STAGEDIR_MOUNT)]
        run_opts = [f'-v "{mp[0]}":"{mp[1]}"' for mp in mount_points]
        if self.workdir:
            run_opts.append(f'-w {self.workdir}')

        run_opts += self.options
        if self.command:
            return (f'docker run --rm {" ".join(run_opts)} '
                    f'{self.image} {self.command}')

        return f'docker run --rm {" ".join(run_opts)} {self.image}'


class Sarus(ContainerPlatform):
    '''Container platform backend for running containers with `Sarus
    <https://sarus.readthedocs.io>`__.'''

    #: Enable MPI support when launching the container.
    #:
    #: :type: boolean
    #: :default: :class:`False`
    with_mpi = fields.TypedField(bool)

    def __init__(self):
        super().__init__()
        self.with_mpi = False
        self._command = 'sarus'

    def emit_prepare_commands(self, stagedir):
        # The format that Sarus uses to call the images is
        # <reposerver>/<user>/<image>:<tag>. If an image was loaded
        # locally from a tar file, the <reposerver> is 'load'.
        if (not self.pull_image or not self.image or
            self.image.startswith('load/')):
            return []
        else:
            return [f'{self._command} pull {self.image}']

    def launch_command(self, stagedir):
        super().launch_command(stagedir)
        mount_points = self.mount_points + [(stagedir, _STAGEDIR_MOUNT)]
        run_opts = [f'--mount=type=bind,source="{mp[0]}",destination="{mp[1]}"'
                    for mp in mount_points]
        if self.with_mpi:
            run_opts.append('--mpi')

        if self.workdir:
            run_opts.append(f'-w {self.workdir}')

        run_opts += self.options
        if self.command:
            return (f'{self._command} run {" ".join(run_opts)} {self.image} '
                    f'{self.command}')

        return f'{self._command} run {" ".join(run_opts)} {self.image}'


class Shifter(Sarus):
    '''Container platform backend for running containers with `Shifter
    <https://www.nersc.gov/research-and-development/user-defined-images/>`__.
    '''

    def __init__(self):
        super().__init__()
        self._command = 'shifter'

    def launch_command(self, stagedir):
        # Temporarily change `workdir`, since Sarus and Shifter have otherwise
        # the same interface
        with util.temp_setattr(self, 'workdir', None):
            return super().launch_command(stagedir)


class Singularity(ContainerPlatform):
    '''Container platform backend for running containers with `Singularity
    <https://sylabs.io/>`__.'''

    #: Enable CUDA support when launching the container.
    #:
    #: :type: boolean
    #: :default: :class:`False`
    with_cuda = fields.TypedField(bool)

    def __init__(self):
        super().__init__()
        self.with_cuda = False
        self._launch_command = 'singularity'

    def emit_prepare_commands(self, stagedir):
        return []

    def launch_command(self, stagedir):
        super().launch_command(stagedir)
        mount_points = self.mount_points + [(stagedir, _STAGEDIR_MOUNT)]
        run_opts = [f'-B"{mp[0]}:{mp[1]}"' for mp in mount_points]
        if self.with_cuda:
            run_opts.append('--nv')

        if self.workdir:
            run_opts.append(f'--pwd {self.workdir}')

        run_opts += self.options
        if self.command:
            return (f'{self._launch_command} exec {" ".join(run_opts)} '
                    f'{self.image} {self.command}')

        return f'{self._launch_command} run {" ".join(run_opts)} {self.image}'


class Apptainer(Singularity):
    '''Container platform backend for running containers with `Apptainer
    <https://apptainer.org/>`__.

    .. versionadded:: 4.0.0

    '''

    def __init__(self):
        super().__init__()
        self._launch_command = 'apptainer'


class ContainerPlatformField(fields.TypedField):
    def __init__(self, *other_types):
        super().__init__(ContainerPlatform, *other_types)

    def __set__(self, obj, value):
        if isinstance(value, str):
            value = ContainerPlatform.create(value)

        super().__set__(obj, value)



# Copyright 2016-2022 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
# ReFrame Project Developers. See the top-level LICENSE file for details.
#
# SPDX-License-Identifier: BSD-3-Clause

import abc
import os
import re

import reframe.core.fields as fields
import reframe.utility.typecheck as typ
from reframe.core.exceptions import BuildSystemError
from reframe.core.meta import RegressionTestMeta


class BuildSystemMeta(RegressionTestMeta, abc.ABCMeta):
    '''Build systems metaclass.'''


class BuildSystem(metaclass=BuildSystemMeta):
    '''The abstract base class of any build system.

    Concrete build systems inherit from this class and must override the
    :func:`emit_build_commands` abstract function.
    '''

    #: The C compiler to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the compiler defined in the current programming environment will be
    #: used.
    #:
    #: :type: :class:`str`
    #: :default: ``''``
    cc  = variable(str, value='')

    #: The C++ compiler to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the compiler defined in the current programming environment will be
    #: used.
    #:
    #: :type: :class:`str`
    #: :default: ``''``
    cxx = variable(str, value='')

    #: The Fortran compiler to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the compiler defined in the current programming environment will be
    #: used.
    #:
    #: :type: :class:`str`
    #: :default: ``''``
    ftn = variable(str, value='')

    #: The CUDA compiler to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the compiler defined in the current programming environment will be
    #: used.
    #:
    #: :type: :class:`str`
    #: :default: ``''``
    nvcc = variable(str, value='')

    #: The C compiler flags to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the corresponding flags defined in the current programming environment
    #: will be used.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    cflags = variable(typ.List[str], value=[])

    #: The preprocessor flags to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the corresponding flags defined in the current programming environment
    #: will be used.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    cppflags = variable(typ.List[str], value=[])

    #: The C++ compiler flags to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the corresponding flags defined in the current programming environment
    #: will be used.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    cxxflags = variable(typ.List[str], value=[])

    #: The Fortran compiler flags to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the corresponding flags defined in the current programming environment
    #: will be used.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    fflags  = variable(typ.List[str], value=[])

    #: The linker flags to be used.
    #: If empty and :attr:`flags_from_environ` is :class:`True`,
    #: the corresponding flags defined in the current programming environment
    #: will be used.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    ldflags = variable(typ.List[str], value=[])

    #: Set compiler and compiler flags from the current programming environment
    #: if not specified otherwise.
    #:
    #: :type: :class:`bool`
    #: :default: :class:`True`
    flags_from_environ = variable(bool, value=True)

    @abc.abstractmethod
    def emit_build_commands(self, environ):
        '''Return the list of commands for building using this build system.

        The build commands, as well as this function, will always be executed
        from the test's stage directory.

        :arg environ: The programming environment for which to emit the build
           instructions.
           The framework passes here the current programming environment.
        :type environ: :class:`reframe.core.environments.ProgEnvironment`
        :raises: :class:`BuildSystemError` in case of errors when generating
          the build instructions.

        .. versionchanged:: 3.5.0
           This function executes from the test stage directory.

        :meta private:

        '''

    def post_build(self, buildjob):
        '''Callback function that the framework will call when the compilation
        is done.

        Build systems may use this information to do some post processing and
        provide additional, build system-specific, functionality to the users.

        This function will always be executed from the test's stage directory.

        .. versionadded:: 3.5.0
        .. versionchanged:: 3.5.2
           The function is executed from the stage directory.

        :meta private:

        '''

    def prepare_cmds(self):
        '''Callback function that the framework will call before run.

        Build systems may use this information to add commands to the run
        script before anything set by the user.

        :meta private:
        '''
        return []

    def _resolve_flags(self, flags, environ):
        _flags = getattr(self, flags)
        if _flags:
            return _flags

        if self.flags_from_environ:
            return getattr(environ, flags)

        return None

    def _cc(self, environ):
        return self._resolve_flags('cc', environ)

    def _cxx(self, environ):
        return self._resolve_flags('cxx', environ)

    def _ftn(self, environ):
        return self._resolve_flags('ftn', environ)

    def _nvcc(self, environ):
        return self._resolve_flags('nvcc', environ)

    def _cppflags(self, environ):
        return self._resolve_flags('cppflags', environ)

    def _cflags(self, environ):
        return self._resolve_flags('cflags', environ)

    def _cxxflags(self, environ):
        return self._resolve_flags('cxxflags', environ)

    def _fflags(self, environ):
        return self._resolve_flags('fflags', environ)

    def _ldflags(self, environ):
        return self._resolve_flags('ldflags', environ)

    def __str__(self):
        return type(self).__name__

    def __rfm_json_encode__(self):
        return str(self)


class Make(BuildSystem):
    '''A build system for compiling codes using ``make``.

    The generated build command has the following form:

    .. code::

      make -j [N] [-f MAKEFILE] [-C SRCDIR] CC="X" CXX="X" FC="X" NVCC="X" CPPFLAGS="X" CFLAGS="X" CXXFLAGS="X" FCFLAGS="X" LDFLAGS="X" OPTIONS

    The compiler and compiler flags variables will only be passed if they are
    not :class:`None`.
    Their value is determined by the corresponding attributes of
    :class:`BuildSystem`.
    If you want to completely disable passing these variables to the ``make``
    invocation, you should make sure not to set any of the correspoding
    attributes and set also the :attr:`BuildSystem.flags_from_environ` flag to
    :class:`False`.
    '''

    #: Append these options to the ``make`` invocation.
    #: This variable is also useful for passing variables or targets to
    #: ``make``.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    options = variable(typ.List[str], value=[])

    #: Instruct build system to use this Makefile.
    #: This option is useful when having non-standard Makefile names.
    #:
    #: :type: :class:`str`
    #: :default: :class:`None`
    makefile = variable(str, type(None), value=None)

    #: The top-level directory of the code.
    #:
    #: This is set automatically by the framework based on the
    #: :attr:`reframe.core.pipeline.RegressionTest.sourcepath` attribute.
    #:
    #: :type: :class:`str`
    #: :default: :class:`None`
    srcdir = variable(str, type(None), value=None)

    #: Limit concurrency for ``make`` jobs.
    #: This attribute controls the ``-j`` option passed to ``make``.
    #: If not :class:`None`, ``make`` will be invoked as ``make -j
    #: max_concurrency``.
    #: Otherwise, it will invoked as ``make -j``.
    #:
    #: :type: integer
    #: :default: ``1``
    #:
    #: .. note::
    #:     .. versionchanged:: 2.19
    #:        The default value is now ``1``
    max_concurrency = variable(int, type(None), value=1)

    def emit_build_commands(self, environ):
        cmd_parts = ['make']
        if self.makefile:
            cmd_parts += [f'-f {self.makefile}']

        if self.srcdir:
            cmd_parts += [f'-C {self.srcdir}']

        cmd_parts += ['-j']
        if self.max_concurrency is not None:
            cmd_parts += [str(self.max_concurrency)]

        cc = self._cc(environ)
        cxx = self._cxx(environ)
        ftn = self._ftn(environ)
        nvcc = self._nvcc(environ)
        cppflags = self._cppflags(environ)
        cflags   = self._cflags(environ)
        cxxflags = self._cxxflags(environ)
        fflags   = self._fflags(environ)
        ldflags  = self._ldflags(environ)
        if cc:
            cmd_parts += [f'CC="{cc}"']

        if cxx:
            cmd_parts += [f'CXX="{cxx}"']

        if ftn:
            cmd_parts += [f'FC="{ftn}"']

        if nvcc:
            cmd_parts += [f'NVCC="{nvcc}"']

        if cppflags:
            flags = ' '.join(cppflags)
            cmd_parts += [f'CPPFLAGS="{flags}"']

        if cflags:
            flags = ' '.join(cflags)
            cmd_parts += [f'CFLAGS="{flags}"']

        if cxxflags:
            flags = ' '.join(cxxflags)
            cmd_parts += [f'CXXFLAGS="{flags}"']

        if fflags:
            flags = ' '.join(fflags)
            cmd_parts += [f'FCFLAGS="{flags}"']

        if ldflags:
            flags = ' '.join(ldflags)
            cmd_parts += [f'LDFLAGS="{flags}"']

        if self.options:
            cmd_parts += self.options

        return [' '.join(cmd_parts)]


class SingleSource(BuildSystem):
    '''A build system for compiling a single source file.

    The generated build command will have the following form:

    .. code::

      COMP CPPFLAGS XFLAGS SRCFILE -o EXEC LDFLAGS

    - ``COMP`` is the required compiler for compiling ``SRCFILE``.
      This build system will automatically detect the programming language of
      the source file and pick the correct compiler.
      See also the :attr:`SingleSource.lang` attribute.
    - ``CPPFLAGS`` are the preprocessor flags and are passed to any compiler.
    - ``XFLAGS`` is any of ``CFLAGS``, ``CXXFLAGS`` or ``FCFLAGS`` depending on
      the programming language of the source file.
    - ``SRCFILE`` is the source file to be compiled.
      This is set up automatically by the framework.
      See also the :attr:`SingleSource.srcfile` attribute.
    - ``EXEC`` is the executable to be generated.
      This is also set automatically by the framework.
      See also the :attr:`SingleSource.executable` attribute.
    - ``LDFLAGS`` are the linker flags.

    For CUDA codes, the language assumed is C++ (for the compilation flags) and
    the compiler used is :attr:`BuildSystem.nvcc`.

    '''

    #: The source file to compile.
    #: This is automatically set by the framework based on the
    #: :attr:`reframe.core.pipeline.RegressionTest.sourcepath` attribute.
    #:
    #: :type: :class:`str` or :class:`None`
    srcfile = variable(str, type(None), value=None)

    #: The executable file to be generated.
    #:
    #: This is set automatically by the framework based on the
    #: :attr:`reframe.core.pipeline.RegressionTest.executable` attribute.
    #:
    #: :type: :class:`str` or :class:`None`
    executable = variable(str, type(None), value=None)

    #: The include path to be used for this compilation.
    #:
    #: All the elements of this list will be appended to the
    #: :attr:`BuildSystem.cppflags`, by prepending to each of them the ``-I``
    #: option.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    include_path = variable(typ.List[str], value=[])

    #: The programming language of the file that needs to be compiled.
    #: If not specified, the build system will try to figure it out
    #: automatically based on the extension of the source file.
    #: The automatically detected extensions are the following:
    #:
    #:   - C: `.c` and `.upc`.
    #:   - C++: `.cc`, `.cp`, `.cxx`, `.cpp`, `.CPP`, `.c++` and `.C`.
    #:   - Fortran: `.f`, `.for`, `.ftn`, `.F`, `.FOR`, `.fpp`, `.FPP`, `.FTN`,
    #:     `.f90`, `.f95`, `.f03`, `.f08`, `.F90`, `.F95`, `.F03` and `.F08`.
    #:   - CUDA: `.cu`.
    #:
    #: :type: :class:`str` or :class:`None`
    lang = variable(str, type(None), value=None)

    def _auto_exec_name(self):
        return os.path.splitext(self.srcfile)[0] + '.x'

    def emit_build_commands(self, environ):
        if not self.srcfile:
            bname = type(self).__name__
            raise BuildSystemError(f'a source file is required when using '
                                   f'the {bname!r} build system')

        cc = self._cc(environ)
        cxx = self._cxx(environ)
        ftn = self._ftn(environ)
        nvcc = self._nvcc(environ)
        cppflags = self._cppflags(environ) or []
        cflags   = self._cflags(environ) or []
        cxxflags = self._cxxflags(environ) or []
        fflags   = self._fflags(environ)  or []
        ldflags  = self._ldflags(environ) or []

        # Adjust cppflags with the include directories
        # NOTE: We do not use the += operator on purpose, be cause we don't
        # want to change the original list passed by the user
        cppflags = cppflags + [*map(lambda d: '-I ' + d, self.include_path)]

        # Generate the executable
        executable = self.executable or self._auto_exec_name()

        # Prepare the compilation command
        lang = self.lang or self._guess_language(self.srcfile)
        cmd_parts = []
        if lang == 'C':
            if not cc:
                raise BuildSystemError('I do not know how to compile a '
                                       'C program')

            cmd_parts += [cc, *cppflags, *cflags, self.srcfile,
                          '-o', executable, *ldflags]
        elif lang == 'C++':
            if not cxx:
                raise BuildSystemError('I do not know how to compile a '
                                       'C++ program')

            cmd_parts += [cxx, *cppflags, *cxxflags, self.srcfile,
                          '-o', executable, *ldflags]
        elif lang == 'Fortran':
            if not ftn:
                raise BuildSystemError('I do not know how to compile a '
                                       'Fortran program')

            cmd_parts += [ftn, *cppflags, *fflags, self.srcfile,
                          '-o', executable, *ldflags]
        elif lang == 'CUDA':
            if not nvcc:
                raise BuildSystemError('I do not know how to compile a '
                                       'CUDA program')

            cmd_parts += [nvcc, *cppflags, *cxxflags, self.srcfile,
                          '-o', executable, *ldflags]
        else:
            raise BuildSystemError(
                f'could not guess language of file: {self.srcfile}'
            )

        return [' '.join(cmd_parts)]

    def _guess_language(self, filename):
        _, ext = os.path.splitext(filename)
        if ext in ['.c', '.upc']:
            return 'C'

        if ext in ['.cc', '.cp', '.cxx', '.cpp', '.CPP', '.c++', '.C']:
            return 'C++'

        if ext in ['.f', '.for', '.ftn', '.F', '.FOR', '.fpp',
                   '.FPP', '.FTN', '.f90', '.f95', '.f03', '.f08',
                   '.F90', '.F95', '.F03', '.F08', '.cuf', '.CUF']:
            return 'Fortran'

        if ext in ['.cu']:
            return 'CUDA'


class ConfigureBasedBuildSystem(BuildSystem):
    '''Abstract base class for configured-based build systems.'''

    #: The top-level directory of the code.
    #:
    #: This is set automatically by the framework based on the
    #: :attr:`reframe.core.pipeline.RegressionTest.sourcepath` attribute.
    #:
    #: :type: :class:`str`
    #: :default: :class:`None`
    srcdir = variable(str, type(None), value=None)

    #: The CMake build directory, where all the generated files will be placed.
    #:
    #: :type: :class:`str`
    #: :default: :class:`None`
    builddir = variable(str, type(None), value=None)

    #: Additional configuration options to be passed to the CMake invocation.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    config_opts = variable(typ.List[str], value=[])

    #: Options to be passed to the subsequent ``make`` invocation.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    make_opts = variable(typ.List[str], value=[])

    #: Same as for the :attr:`Make` build system.
    #:
    #: :type: integer
    #: :default: ``1``
    max_concurrency = variable(int, type(None), value=1)


class CMake(ConfigureBasedBuildSystem):
    '''A build system for compiling CMake-based projects.

    This build system will emit the following commands:

    1. Create a build directory if :attr:`builddir` is not :class:`None` and
       change to it.
    2. Invoke ``cmake`` to configure the project by setting the corresponding
       CMake flags for compilers and compiler flags.
    3. Issue ``make`` to compile the code.
    '''

    def _combine_flags(self, cppflags, xflags):
        if not cppflags:
            return xflags

        ret = list(cppflags)
        if xflags:
            ret += xflags

        return ret

    def emit_build_commands(self, environ):
        prepare_cmd = []
        if self.srcdir:
            prepare_cmd += [f'cd {self.srcdir}']

        if self.builddir:
            prepare_cmd += [f'mkdir -p {self.builddir}',
                            f'cd {self.builddir}']

        cmake_cmd = ['cmake']
        cc = self._cc(environ)
        cxx = self._cxx(environ)
        ftn = self._ftn(environ)
        nvcc = self._nvcc(environ)
        cppflags = self._cppflags(environ)
        cflags   = self._combine_flags(cppflags, self._cflags(environ))
        cxxflags = self._combine_flags(cppflags, self._cxxflags(environ))
        fflags   = self._combine_flags(cppflags, self._fflags(environ))
        ldflags  = self._ldflags(environ)
        if cc:
            cmake_cmd += [f'-DCMAKE_C_COMPILER="{cc}"']

        if cxx:
            cmake_cmd += [f'-DCMAKE_CXX_COMPILER="{cxx}"']

        if ftn:
            cmake_cmd += [f'-DCMAKE_Fortran_COMPILER="{ftn}"']

        if nvcc:
            cmake_cmd += [f'-DCMAKE_CUDA_COMPILER="{nvcc}"']

        if cflags:
            flags = ' '.join(cflags)
            cmake_cmd += [f'-DCMAKE_C_FLAGS="{flags}"']

        if cxxflags:
            flags = ' '.join(cxxflags)
            cmake_cmd += [f'-DCMAKE_CXX_FLAGS="{flags}"']

        if fflags:
            flags = ' '.join(fflags)
            cmake_cmd += [f'-DCMAKE_Fortran_FLAGS="{flags}"']

        if ldflags:
            flags = ' '.join(ldflags)
            cmake_cmd += [f'-DCMAKE_EXE_LINKER_FLAGS="{flags}"']

        if self.config_opts:
            cmake_cmd += self.config_opts

        if self.builddir:
            cmake_cmd += [os.path.relpath('.', self.builddir)]
        else:
            cmake_cmd += ['.']

        make_cmd = ['make -j']
        if self.max_concurrency is not None:
            make_cmd += [str(self.max_concurrency)]

        if self.make_opts:
            make_cmd += self.make_opts

        return prepare_cmd + [' '.join(cmake_cmd), ' '.join(make_cmd)]


class Autotools(ConfigureBasedBuildSystem):
    '''A build system for compiling Autotools-based projects.

    This build system will emit the following commands:

    1. Create a build directory if :attr:`builddir` is not :class:`None` and
       change to it.
    2. Invoke ``configure`` to configure the project by setting the
       corresponding flags for compilers and compiler flags.
    3. Issue ``make`` to compile the code.
    '''

    #: The directory of the configure script.
    #:
    #: This can be changed to do an out of source build without copying the
    #: entire source tree.
    #:
    #: :type: :class:`str`
    #: :default: ``'.'``
    configuredir = variable(str, value='.')

    def emit_build_commands(self, environ):
        prepare_cmd = []
        if self.srcdir:
            prepare_cmd += [f'cd {self.srcdir}']

        if self.builddir:
            prepare_cmd += [f'mkdir -p {self.builddir}',
                            f'cd {self.builddir}']

        if self.builddir:
            configure_cmd = [os.path.join(
                os.path.relpath(self.configuredir, self.builddir), 'configure'
            )]
        else:
            configure_cmd = [os.path.join(self.configuredir, 'configure')]

        cc = self._cc(environ)
        cxx = self._cxx(environ)
        ftn = self._ftn(environ)
        cppflags = self._cppflags(environ)
        cflags   = self._cflags(environ)
        cxxflags = self._cxxflags(environ)
        fflags   = self._fflags(environ)
        ldflags  = self._ldflags(environ)
        if cc:
            configure_cmd += [f'CC="{cc}"']

        if cxx:
            configure_cmd += [f'CXX="{cxx}"']

        if ftn:
            configure_cmd += [f'FC="{ftn}"']

        if cppflags:
            flags = ' '.join(cppflags)
            configure_cmd += [f'CPPFLAGS="{flags}"']

        if cflags:
            flags = ' '.join(cflags)
            configure_cmd += [f'CFLAGS="{flags}"']

        if cxxflags:
            flags = ' '.join(cxxflags)
            configure_cmd += [f'CXXFLAGS="{flags}"']

        if fflags:
            flags = ' '.join(fflags)
            configure_cmd += [f'FCFLAGS="{flags}"']

        if ldflags:
            flags = ' '.join(ldflags)
            configure_cmd += [f'LDFLAGS="{flags}"']

        if self.config_opts:
            configure_cmd += self.config_opts

        make_cmd = ['make -j']
        if self.max_concurrency is not None:
            make_cmd += [str(self.max_concurrency)]

        if self.make_opts:
            make_cmd += self.make_opts

        return prepare_cmd + [' '.join(configure_cmd), ' '.join(make_cmd)]


class EasyBuild(BuildSystem):
    '''A build system for building test code using `EasyBuild
    <https://easybuild.io/>`__.

    ReFrame will use EasyBuild to build and install the code in the test's
    stage directory by default. ReFrame uses environment variables to
    configure EasyBuild for running, so users can pass additional options to
    the ``eb`` command and modify the default behaviour.

    .. versionadded:: 3.5.0

    '''

    #: The list of easyconfig files to build and install.
    #: This field is required.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    easyconfigs = variable(typ.List[str], value=[])

    #: Options to pass to the ``eb`` command.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    options = variable(typ.List[str], value=[])

    #: Instruct EasyBuild to emit a package for the built software.
    #: This will essentially pass the ``--package`` option to ``eb``.
    #:
    #: :type: :class:`bool`
    #: :default: ``False``
    emit_package = variable(bool, value=False)

    #: Options controlling the package creation from EasyBuild.
    #: For each key/value pair of this dictionary, ReFrame will pass
    #: ``--package-{key}={val}`` to the EasyBuild invocation.
    #:
    #: :type: :class:`Dict[str, str]`
    #: :default: ``{}``
    package_opts = variable(typ.Dict[str, str], value={})

    #: Default prefix for the EasyBuild installation.
    #:
    #: Relative paths will be appended to the stage directory of the test.
    #: ReFrame will set the following environment variables before running
    #: EasyBuild.
    #:
    #: .. code-block:: bash
    #:
    #:    export EASYBUILD_BUILDPATH={prefix}/build
    #:    export EASYBUILD_INSTALLPATH={prefix}
    #:    export EASYBUILD_PREFIX={prefix}
    #:    export EASYBUILD_SOURCEPATH={prefix}
    #:
    #: Users can change these defaults by passing specific options to the
    #: ``eb`` command.
    #:
    #: :type: :class:`str`
    #: :default: ``easybuild``
    prefix = variable(str, value='easybuild')

    def __init__(self):
        self._eb_modules = []

    def emit_build_commands(self, environ):
        if not self.easyconfigs:
            raise BuildSystemError(f"'easyconfigs' must not be empty")

        easyconfigs = ' '.join(self.easyconfigs)
        if self.emit_package:
            self.options.append('--package')
            for key, val in self.package_opts.items():
                self.options.append(f'--package-{key}={val}')

        prefix = os.path.join(os.getcwd(), self.prefix)
        options = ' '.join(self.options)
        return [f'export EASYBUILD_BUILDPATH={prefix}/build',
                f'export EASYBUILD_INSTALLPATH={prefix}',
                f'export EASYBUILD_PREFIX={prefix}',
                f'export EASYBUILD_SOURCEPATH={prefix}',
                f'eb {easyconfigs} {options}']

    def post_build(self, buildjob):
        # Store the modules generated by EasyBuild

        modulesdir = os.path.join(os.getcwd(), self.prefix,
                                  'modules', 'all')
        with open(buildjob.stdout) as fp:
            out = fp.read()

        self._eb_modules = [
            {'name': m, 'collection': False, 'path': modulesdir}
            for m in re.findall(r'building and installing (\S+)...', out)
        ]

    @property
    def generated_modules(self):
        return self._eb_modules


class Spack(BuildSystem):
    '''A build system for building test code using `Spack
    <https://spack.io/>`__.

    ReFrame will use a user-provided Spack environment in order to build and
    test a set of specs.

    .. versionadded:: 3.6.1

    '''

    #: The Spack environment to use for building this test.
    #:
    #: ReFrame will activate and install this environment.
    #: This environment will also be used to run the test.
    #:
    #: .. code-block:: bash
    #:
    #:    spack env activate -V -d <environment directory>
    #:
    #: ReFrame looks for environments in the test's
    #: :attr:`~reframe.core.pipeline.RegressionTest.sourcesdir`.
    #:
    #: If this field is `None`, the default, the environment name will
    #: be automatically set to `rfm_spack_env`.
    #:
    #: :type: :class:`str` or :class:`None`
    #: :default: :class:`None`
    #:
    #: .. note::
    #:     .. versionchanged:: 3.7.3
    #:        The field is no longer required and the Spack environment will be
    #:        automatically created if not provided.
    environment = variable(typ.Str[r'\S+'], type(None), value=None)

    #: The directory where Spack will install the packages requested by this
    #: test.
    #:
    #: After activating the Spack environment, ReFrame will set the
    #: `install_tree` Spack configuration in the given environment with the
    #: following command:
    #:
    #: .. code-block:: bash
    #:
    #:    spack config add "config:install_tree:root:<install tree>"
    #:
    #: Relative paths are resolved against the test's stage directory.  If this
    #: field and the Spack environment are both `None`, the default, the
    #: install directory will be automatically set to `opt/spack`.  If this
    #: field `None` but the Spack environment is not, then `install_tree` will
    #: not be set automatically and the install tree of the given environment
    #: will not be overridden.
    #:
    #: :type: :class:`str` or :class:`None`
    #: :default: :class:`None`
    #:
    #: .. versionadded:: 3.7.3
    install_tree = variable(typ.Str[r'\S+'], type(None), value=None)

    #: A list of additional specs to build and install within the given
    #: environment.
    #:
    #: ReFrame will add the specs to the active environment by emititing the
    #: following command:
    #:
    #: .. code-block:: bash
    #:
    #:    spack add spec1 spec2 ... specN
    #:
    #: If no spec is passed, ReFrame will simply install what is prescribed by
    #: the environment.
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    specs = variable(typ.List[str], value=[])

    #: Emit the necessary ``spack load`` commands before running the test.
    #:
    #: :type: :class:`bool`
    #: :default: :obj:`True`
    emit_load_cmds = variable(bool, value=True)

    #: Options to pass to ``spack install``
    #:
    #: :type: :class:`List[str]`
    #: :default: ``[]``
    install_opts = variable(typ.List[str], value=[])

    def __init__(self):
        # Set to True if the environment was auto-generated
        self._auto_env = False

    def emit_build_commands(self, environ):
        ret = self._create_env_cmds()

        if self._auto_env:
            install_tree = self.install_tree or 'opt/spack'
            ret.append(f'spack -e {self.environment} config add '
                       f'"config:install_tree:root:{install_tree}"')

        if self.specs:
            specs_str = ' '.join(self.specs)
            ret.append(f'spack -e {self.environment} add {specs_str}')

        install_cmd = f'spack -e {self.environment} install'
        if self.install_opts:
            install_cmd += ' ' + ' '.join(self.install_opts)

        ret.append(install_cmd)
        return ret

    def _create_env_cmds(self):
        if self.environment:
            return []

        self.environment = 'rfm_spack_env'
        self._auto_env = True
        return [f'spack env create -d {self.environment}']

    def prepare_cmds(self):
        cmds = self._create_env_cmds()
        if self.specs and self.emit_load_cmds:
            cmds.append(
                f'eval `spack -e {self.environment} load '
                f'--sh {" ".join(self.specs)}`'
            )

        return cmds


class CustomBuild(BuildSystem):
    '''Custom build system.

    This build system backend allows users to use custom build scripts to
    build the test code. It does not do any interpretation of the current test
    environment and it simply runs the supplied :attr:`commands`.

    Users should use this build system with caution, because environment
    management, reproducibility and any potential side effects are all
    controlled by the user's custom build system.

    .. versionadded:: 3.11.0
    '''

    #: The commands to run for building the test code.
    #:
    #: :type: :class:`List[str]`
    commands = variable(typ.List[str])

    def emit_build_commands(self, environ):
        return self.commands


class BuildSystemField(fields.TypedField):
    def __init__(self, fieldname, *other_types):
        super().__init__(fieldname, BuildSystem, *other_types)

    def __set__(self, obj, value):
        if isinstance(value, str):
            try:
                value = globals()[value]()
            except KeyError:
                raise ValueError(f'unknown build system: {value}') from None

        super().__set__(obj, value)
